
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <pthread.h>
#include <semaphore.h>
#include "globdef.h"
#include "uidef.h"
#include "sigdef.h"
#include "fft1def.h"
#include "fft2def.h"
#include "fft3def.h"
#include "conf.h"
#include "lconf.h"
#include "screendef.h"
#include "sdrdef.h"
#include "vernr.h"
#include "thrdef.h"
#include "hwaredef.h"
#include "rusage.h"
#include "ldef.h"
#include "txdef.h"

#define ABOVE_MAX_SPEED 768000
#define MAX_COLUMN 80


#define SDR14_DEVICE_DRIVER "/dev/linrad_ft245"
#if SND_DEV == 1
#define DEVNAME_BASE "/dev/dsp"
#define DEVNAME_SIZE 12
#else
#define DEVNAME_BASE "/dev/sound/dsp"
#define DEVNAME_SIZE 18
#endif
#define DEVFLAG_R 1
#define DEVFLAG_W 2
#define DEVFLAG_RW 4
char dev_name[sizeof(DEVNAME_BASE)+2];
int devmodes[3]={O_RDONLY,O_WRONLY,O_RDWR};
char *devmode_txt[3]={"RDONLY","WRONLY","RDWR"};
#define DEVMODE_RDONLY 0
#define DEVMODE_WRONLY 1
#define DEVMODE_RDWR 2

#define RXAD 0
#define RXDA 1
#define TXDA 2
#define TXAD 3
#define MAX_IOTEST 3

void thread_test_rxad(void);
void thread_test_rxda(void);
void thread_test_txda(void);

int soundcard_test_block_count[MAX_IOTEST];
int soundcard_test_cmd_flag[MAX_IOTEST];
pthread_t thread_identifier_soundcard_test[MAX_IOTEST];
ROUTINE thread_iotest[MAX_IOTEST]={thread_test_rxad,     //0
                                   thread_test_rxda,     //1
                                   thread_test_txda};    //2

int dev_flag[MAX_DEVNAMES];
int dev_rd_channels[MAX_DEVNAMES];
int dev_max_rd_speed[MAX_DEVNAMES];
int dev_min_rd_speed[MAX_DEVNAMES];
int dev_rd_bytes[MAX_DEVNAMES];
int dev_wr_channels[MAX_DEVNAMES];
int dev_max_wr_speed[MAX_DEVNAMES];
int dev_min_wr_speed[MAX_DEVNAMES];
int dev_wr_bytes[MAX_DEVNAMES];
char blanks[60];

char *rx_logfile_name="soundboard_rx_init.log";

int low_speeds[MAX_LOWSPEED]={1,5000,6000,8000,11025,16000,
                              22050,24000,48000,96000};

#ifdef ALSA_PRESENT
// alsa definitions start
#include <alsa/asoundlib.h>

snd_pcm_t *rx_ad_handle,*rx_da_handle;
snd_pcm_t *tx_ad_handle,*tx_da_handle;
snd_pcm_uframes_t rx_da_period_size;
snd_pcm_uframes_t tx_da_period_size;
snd_pcm_uframes_t rx_ad_frames, rx_da_frames;
snd_pcm_uframes_t tx_ad_frames, tx_da_frames;
int rx_da_bytes_per_frame;
int tx_da_bytes_per_frame;
snd_pcm_uframes_t rx_da_buffer_size;
snd_pcm_uframes_t tx_da_buffer_size;
int tx_read_fragments;

int  alsa_dev_seq_nmbr;
char alsa_dev_soundcard_name [64];
char alsa_dev_name [64];
char alsa_dev_hw_pcm_name [10];
char alsa_dev_plughw_pcm_name [10];
char sndout_plughw_pcm_name [10];
char sndin_plughw_pcm_name [10];
int  alsa_dev_min_rate;
int  alsa_dev_max_rate;
int  alsa_dev_min_channels;
int  alsa_dev_max_channels;
int  alsa_dev_min_bytes;
int  alsa_dev_max_bytes;

int  alsa_get_dev_id (int n)
{
//This routine returns  alsa_dev_hw_pcm_name
//               	alsa_dev_plughw_pcm_name
//			alsa_dev_soundcard_name 
//			alsa_dev_name
//for a valid  alsa_dev_seq_nmbr  on input.
//Input value must be positive and less than  17   
//If error, returncode is negative.
int err, idx, dev, dev_seq_nmbr, retcode;
snd_ctl_t *handle;
snd_ctl_card_info_t *info;
snd_pcm_info_t *pcminfo;
char str[128];
retcode=-1;
alsa_dev_seq_nmbr=0;
sprintf(alsa_dev_hw_pcm_name ," ");
sprintf(alsa_dev_plughw_pcm_name," ");
sprintf(alsa_dev_soundcard_name," ");
sprintf(alsa_dev_name," ");
idx=-1;
dev_seq_nmbr=0;
if(n < 0) goto exit_2;
if(n > 16) goto exit_2;
snd_ctl_card_info_malloc(&info);
snd_pcm_info_malloc(&pcminfo);
while (1) 
  {
  if((err=snd_card_next(&idx)) < 0) 
    {
    SNDLOG"alsa_get_dev_id:snd_card_next error: %s\n", snd_strerror(err));
    break;
    }
  if(idx < 0)	break;               // no more cards available 
  sprintf(str, "hw:CARD=%i", idx);
  SNDLOG"Testing %s\n",str);
  if((err=snd_ctl_open( &handle, str, 0)) < 0) 
    {
    SNDLOG"alsa_get_dev_id:snd_ctl_open error: %s\n", snd_strerror(err));
    }
  if((err=snd_ctl_card_info(handle, info)) < 0) 
    {
    SNDLOG"alsa_get_dev_id:HW info error: %s\n", snd_strerror(err));
    }
  dev=-1;
  while (1) 
    {
    if((err=snd_ctl_pcm_next_device(handle, &dev)) < 0) 
      {
      SNDLOG"alsa_get_dev_id:PCM next device error: %s\n", snd_strerror(err));
      retcode=-1;
      break;
      }
    if(dev < 0)  break;       // no more devices available 
    snd_pcm_info_set_device(pcminfo, dev);
    snd_pcm_info_set_subdevice(pcminfo, 0);
    if((err=snd_ctl_pcm_info(handle, pcminfo)) < 0) 
      {
      SNDLOG"alsa_get_dev_id:PCM info error: %s\n", snd_strerror(err));
      continue;
      }
    if(dev_seq_nmbr == n) 
      {
      alsa_dev_seq_nmbr=dev_seq_nmbr;
      sprintf(alsa_dev_hw_pcm_name ,"hw:%i,%i", idx,dev);
      sprintf(alsa_dev_plughw_pcm_name,"plughw:%i,%i", idx,dev);
      sprintf(alsa_dev_soundcard_name," %s", snd_ctl_card_info_get_longname(info));
      sprintf(alsa_dev_name, "%s", snd_pcm_info_get_name(pcminfo));
      retcode=0;
      goto exit_1;
      }
    dev_seq_nmbr= dev_seq_nmbr+1;
    }
  }
exit_1:;
snd_ctl_close(handle);
snd_ctl_card_info_free(info);
snd_pcm_info_free(pcminfo);
exit_2:;	
return retcode;
}

int  alsa_get_dev_native_capabilities(int n,int stream)
{
snd_pcm_t *wk_handle;
unsigned int val;
snd_pcm_hw_params_t *hw_params;
// This routine returns the native hw capabilities of a given 
// soundcard/device in either capture or playback mode (stream).
// This is done by opening the device with a pcm_name=hw:n,n  and 
// querying hw_params
// If error, returncode is negative.
int err, dir, retcode;
dir=0;
retcode=-1;
alsa_dev_min_rate=0;
alsa_dev_max_rate=0;
alsa_dev_min_channels=0;
alsa_dev_max_channels=0;
alsa_dev_min_bytes=0;
alsa_dev_max_bytes=0;
if((err=alsa_get_dev_id (n)) < 0 ) goto exit_1;;
if((err=snd_pcm_open(&wk_handle, alsa_dev_hw_pcm_name, stream, 0)) < 0) goto exit_1;
if((err=snd_pcm_hw_params_malloc (&hw_params)) < 0) goto exit_2;
if((err=snd_pcm_hw_params_any(wk_handle, hw_params)) < 0) goto exit_3;
// get sampling rate
if((err= snd_pcm_hw_params_get_rate_max (hw_params,	&val,&dir)) < 0) goto exit_3;
alsa_dev_max_rate=val;
if((err= snd_pcm_hw_params_get_rate_min (hw_params,	&val,&dir)) < 0)  goto exit_3;
alsa_dev_min_rate=val;
// get # channels
if((err= snd_pcm_hw_params_get_channels_max(hw_params, &val)) < 0)  goto exit_3;
alsa_dev_max_channels=val;
if((err= snd_pcm_hw_params_get_channels_min(hw_params, &val)) < 0)  goto exit_3;
alsa_dev_min_channels=val;
// get  # of bytes  
if((err=snd_pcm_hw_params_test_format (wk_handle, hw_params,SND_PCM_FORMAT_U8)) == 0)
  {
  alsa_dev_min_bytes=1;
  }
else 
  {
  if((err=snd_pcm_hw_params_test_format (wk_handle, hw_params,SND_PCM_FORMAT_S16_LE)) == 0)
    {
    alsa_dev_min_bytes=2;
    }
  else 
    {
    if((err=snd_pcm_hw_params_test_format (wk_handle, hw_params,SND_PCM_FORMAT_S32_LE)) == 0)
      {
      alsa_dev_min_bytes=4;
      }
    else goto exit_3;
    }
  }
if((err=snd_pcm_hw_params_test_format (wk_handle, hw_params,SND_PCM_FORMAT_S32_LE)) == 0)
  {
  alsa_dev_max_bytes=4;
  }
else 
  {
  if((err=snd_pcm_hw_params_test_format (wk_handle, hw_params,SND_PCM_FORMAT_S16_LE)) == 0)
    {
    alsa_dev_max_bytes=2;
    }
  else 
    {
    if((err=snd_pcm_hw_params_test_format (wk_handle, hw_params,SND_PCM_FORMAT_U8)) == 0)
      {
      alsa_dev_max_bytes=1;
      }
    else goto exit_3;
    }
  }
retcode=0;
exit_3:; 
snd_pcm_hw_params_free (hw_params);
exit_2:;
snd_pcm_close(wk_handle);	
exit_1:;
return retcode;
}


int select_alsadev(int open_mode, int *line)
{
char s[200];
int i, err;
line[0]++;
sprintf(s,"device                                                      device      sample-rate  channels  bytes");
SNDLOG"%s\n",s);
lir_text(0,line[0],s);
line[0]++;
sprintf(s,"seq-no            soundcard-name / device-name             pcm-name      min / max   min/max  min/max");
SNDLOG"%s\n",s);
lir_text(0,line[0],s);
line[0]++;
sprintf(s,"------ --------------------------------------------------  ----------   -----------  -------  ------- " );
SNDLOG"%s\n",s);
lir_text(0,line[0],s);
line[0]++;
for(i=0; i<17; i++) 
  {	
  err=alsa_get_dev_id(i);
  if(err == 0)
    {
    err=alsa_get_dev_native_capabilities(i, open_mode);
    if(err == 0)
      {                  
      line[0]++;
      sprintf(s,"%i", alsa_dev_seq_nmbr);
      lir_text(2,line[0],s);
      sprintf(s,"%s", alsa_dev_soundcard_name);
      lir_text(6,line[0],s);
      sprintf(s,"%s", alsa_dev_name );
      lir_text(7,line[0]+1,s);
      sprintf(s,"%s", alsa_dev_plughw_pcm_name);
      lir_text(59,line[0],s);
      sprintf(s,"%i/%i", alsa_dev_min_rate, alsa_dev_max_rate);
      lir_text(72,line[0],s);
      sprintf(s,"%i/%i ", alsa_dev_min_channels, alsa_dev_max_channels);
      lir_text(86,line[0],s);
      sprintf(s,"%i/%i ", alsa_dev_min_bytes, alsa_dev_max_bytes);
      lir_text(96,line[0],s);
      sprintf(s,"     %i   %s / %s    %s   %i/%i   %i/%i   %i/%i ", 
             alsa_dev_seq_nmbr, alsa_dev_soundcard_name, alsa_dev_name,
             alsa_dev_plughw_pcm_name, alsa_dev_min_rate, alsa_dev_max_rate, 
             alsa_dev_min_channels, alsa_dev_max_channels, alsa_dev_min_bytes, 
                                                          alsa_dev_max_bytes );
      line[0]++;
      SNDLOG"%s\n",s);
      }
    }
  }
sprintf(s,"Enter selected device-seq-nmbr >" );
SNDLOG"%s\n",s);  
line[0]=line[0]+2;
again:;
lir_text(0,line[0],s);
i=lir_get_integer(33, line[0], 2, 0,16);
if(kill_all_flag)return -1;
sprintf(s,"User selected  %i",i);
SNDLOG"%s\n",s);  
err=alsa_get_dev_id (i);
if(err < 0)
  {
  lir_text(37,line[0],"invalid device-seq-nmbr, try again");
  goto again;
  }
err=alsa_get_dev_native_capabilities(i,open_mode);
lir_text(37,line[0],blanks);
line[0]++;
return i;
}

int alsa_getospace_rxda (void)
{
snd_pcm_sframes_t frame_delay;
snd_pcm_sframes_t frame_avail;

//This routine mimics the OSS call ioctl(rx_audio_out,SNDCTL_DSP_GETOSPACE, &rx_da_info)  
//On error, returncode is negative.

// A fragment in OSS is called a period in ALSA
// Unit of measurement in OSS for fragments is bytes
// Unit of measurement in ALSA  for periods and buffersizes is frames
// frame == sample
// The size in bytes of a frame (or sample)  is=#channels * #bytes for the selected format

//typedef struct audio_buf_info  in OSS
//	int fragments;     / # of available fragments (partially usend ones not counted)/
//	int fragstotal;    / Total # of fragments allocated in buffer/
//	int fragsize;      / Size of a fragment in bytes /
//	int bytes;         / Available space in bytes (includes partially used fragments) /

int err, retcode;
retcode=0;
rx_da_info.fragments=0;
rx_da_info.fragstotal=0;
rx_da_info.fragsize=0; 
rx_da_info.bytes=0;
//Synchronize stream position with hardware 
err=snd_pcm_hwsync  ( rx_da_handle);
if(err != 0)
  {
  if(err == -EPIPE)       //       underrun 
    {
    err=snd_pcm_prepare(rx_da_handle);
    if(err < 0)
      {
      retcode=1101;
      goto error_exit;
      }
    }
  else 
    {
    if(err == -ESTRPIPE)   //       suspend 
      {
      while ((err=snd_pcm_resume(rx_da_handle)) == -EAGAIN) sleep(1);    // * wait until the suspend flag is released * !!!! ????
      if(err < 0)
        {
	err=snd_pcm_prepare(rx_da_handle);
	if(err < 0)
	  {
          retcode=1104;
	  goto error_exit;
	  }
	}
      }
    else if(err < 0)
      {
      retcode=1107;
      goto error_exit;
      }
    }
  }		
//Get # of frames available for read or write.
frame_avail=snd_pcm_avail_update(rx_da_handle); 
//Get # of frames written  awaiting processing:
frame_delay=rx_da_buffer_size-frame_avail; 
rx_da_info.fragments=(int)frame_avail/(int) rx_da_period_size;
rx_da_info.fragstotal=((int)rx_da_buffer_size)/(int) rx_da_period_size;
rx_da_info.fragsize=(int)rx_da_period_size*rx_da_bytes_per_frame; 
rx_da_info.bytes=(int)frame_avail*rx_da_bytes_per_frame;
error_exit:;
return retcode ;
}

int alsa_getospace_txda (void)
{
snd_pcm_sframes_t frame_delay;
snd_pcm_sframes_t frame_avail;
//This routine mimics the OSS call ioctl(rx_audio_out,
//                                        SNDCTL_DSP_GETOSPACE, &tx_da_info)  
//On error, returncode is negative.
// A fragment in OSS is called a period in ALSA
// Unit of measurement in OSS for fragments is bytes
// Unit of measurement in ALSA  for periods and buffersizes is frames
// frame == sample
// The size in bytes of a frame (or sample)  is=#channels * #bytes for the selected format 
//typedef struct audio_buf_info  in OSS
//	int fragments;     / # of available fragments (partially usend ones not counted)/
//	int fragstotal;    / Total # of fragments allocated in buffer/
//	int fragsize;      / Size of a fragment in bytes /
//	int bytes;         / Available space in bytes (includes partially used fragments) /

int err, retcode;
retcode =0;
tx_da_info.fragments=0;
tx_da_info.fragstotal=0;
tx_da_info.fragsize=0; 
tx_da_info.bytes=0;
//Synchronize stream position with hardware 
err=snd_pcm_hwsync  ( tx_da_handle);
if(err != 0)
  {
  if(err == -EPIPE)       //       underrun 
    {
    err=snd_pcm_prepare(tx_da_handle);
    if(err < 0)
      {
      retcode=1101;
      goto error_exit;
      }
    }
  else 
    {
    if(err == -ESTRPIPE)   //       suspend 
      {
      while ((err=snd_pcm_resume(tx_da_handle)) == -EAGAIN) sleep(1);    // * wait until the suspend flag is released * !!!! ????
      if(err < 0)
        {
	err=snd_pcm_prepare(tx_da_handle);
	if(err < 0)
	  {
          retcode=1104;
	  goto error_exit;
	  }
	}
      }
    else if(err < 0)
      {
      retcode=1107;
      goto error_exit;
      }
    }
  }		
//Get # of frames available for read or write.
frame_avail=snd_pcm_avail_update(tx_da_handle); 
//Get # of frames written  awaiting processing:
frame_delay=tx_da_buffer_size-frame_avail; 
tx_da_info.fragments=(int)frame_avail/(int) tx_da_period_size;
tx_da_info.fragstotal=((int)tx_da_buffer_size)/(int) tx_da_period_size;
tx_da_info.fragsize=(int)tx_da_period_size*tx_da_bytes_per_frame; 
tx_da_info.bytes=(int)frame_avail*tx_da_bytes_per_frame;
error_exit:;
return retcode;
}
#endif


void lir_sdr14_write(void *s, int bytes)
{
write(sdr,s,bytes);
}

int lir_sdr14_read(void *s,int bytes)
{
return read(sdr, s, bytes);
}

void open_sdr14(void)
{
SNDLOG"Trying to open %s\n",SDR14_DEVICE_DRIVER);
sdr=open( SDR14_DEVICE_DRIVER ,O_RDWR );
}

void close_sdr14(void)
{
int i;
i=fcntl(sdr, F_GETFD);
i|=O_NONBLOCK;
fcntl(sdr,F_SETFD,i);
lir_sleep(10000);
close(sdr);
sdr=-1;
}

void lir_tx_adread(char *buf)
{    
int nread, err;
#ifdef ALSA_PRESENT
if(ui.use_alsa)
  {
  err=snd_pcm_readi(tx_ad_handle, buf, tx_ad_frames);
  if((unsigned int)err != tx_ad_frames)
    {
    if(err == -EPIPE) //         EPIPE means overrun 
      { 
      err=snd_pcm_prepare(tx_ad_handle);
      if(err < 0)lirerr(1108);  
      }
    else 
      {
      if(err < 0) 
        {
        lirerr(1113);
        }
      else
        {
        lirerr(1116);
        }
      }
    }
  }
else
#endif
  {
  nread=read(tx_audio_in, buf, tx_read_bytes); 
  if(nread != tx_read_bytes)lirerr(1283);
  }
}  

void thread_rx_adinput(void)
{
double dt1, read_start_time,total_reads;
double total_time1, total_time2;
short int *rxin_isho2;
int *rxin_int, *rxin_int2;
int i, err;
int timing_loop_counter,timing_loop_counter_max,initial_skip_flag;
int screen_loop_counter,screen_loop_counter_max;
#if (RUSAGE_OLD == 1)
int wlcnt,wlcnt_max;
int local_reset;
float t1,t3;
double cpu_time1;
double cpu_time2;
total_time1=current_time();
cpu_time1=lir_get_cpu_time();
wlcnt_max=2*interrupt_rate;
wlcnt=wlcnt_max;
local_reset=workload_reset_flag;
#endif
rxinput_local_px=timf1p_pa;
rxin_local_workload_reset=workload_reset_flag;
timing_loop_counter_max=interrupt_rate;
timing_loop_counter=2.5*timing_loop_counter_max;
screen_loop_counter_max=0.1*interrupt_rate;
if(screen_loop_counter_max==0)screen_loop_counter_max=1;
screen_loop_counter=interrupt_rate;
open_rx_sndin();
if(kill_all_flag) goto rx_adin_init_error;
if(thread_command_flag[THREAD_SCREEN] != THRFLAG_NOT_ACTIVE)
 {
  while(thread_status_flag[THREAD_SCREEN]!=THRFLAG_ACTIVE &&
        thread_status_flag[THREAD_SCREEN]!=THRFLAG_IDLE &&
        thread_status_flag[THREAD_SCREEN]!=THRFLAG_SEM_WAIT)
    {
    if(thread_command_flag[THREAD_RX_ADINPUT] == 
                                           THRFLAG_KILL)goto rxadin_error_exit;
    lir_sleep(10000);
    }
  }
total_time1=current_time();
read_start_time=total_time1;
total_reads=0;
initial_skip_flag=1;
thread_status_flag[THREAD_RX_ADINPUT]=THRFLAG_ACTIVE;
while(thread_command_flag[THREAD_RX_ADINPUT] == THRFLAG_ACTIVE)
  {
#if (RUSAGE_OLD==1)
  wlcnt--;
  if(wlcnt==0)
    {
    wlcnt=wlcnt_max;
    total_time2=current_time();
    cpu_time2=lir_get_cpu_time();
    t3=100/(total_time2-total_time1);
    t1=(cpu_time2-cpu_time1)*t3;
    if(t1<0)t1=0;
    workloads[THREAD_RX_ADINPUT]=t1;
    if(local_reset != workload_reset_flag)
      {
      local_reset=workload_reset_flag;
      total_time1=total_time2;
      cpu_time1=cpu_time2;
      }
    }
#endif
  timing_loop_counter--;
  if(timing_loop_counter == 0)
    {
    timing_loop_counter=timing_loop_counter_max;
    total_time2=current_time();
    if(initial_skip_flag != 0)
      {
      read_start_time=total_time2;
      total_reads=0;
      initial_skip_flag=0;
      }
    else
      {
      total_reads+=timing_loop_counter_max;
      dt1=total_time2-read_start_time;
      measured_ad_speed=total_reads*ad_read_fragments/dt1;
      }
    }
// Here we post to the screen routine every 0.1 second.
  screen_loop_counter--;
  if(screen_loop_counter == 0)
    {
    screen_loop_counter=screen_loop_counter_max;
    lir_sem_post(SEM_SCREEN);
    }
  rxin_isho=(void*)(&timf1_char[timf1p_pa]);
  timf1p_pa=(timf1p_pa+rx_read_bytes)&timf1_bytemask;
  if(ui.rx_addev_no < 256)
    {
#ifdef ALSA_PRESENT
    if(ui.use_alsa)
      {
      err=snd_pcm_readi(rx_ad_handle,rxin_isho, rx_ad_frames);
      if((unsigned int)err != rx_ad_frames)
	{
        if(err == -EPIPE) //         EPIPE means overrun 
	  { 
	  err=snd_pcm_prepare(rx_ad_handle);
          if(err < 0)lirerr(1108);  
	  }
        else 
          {
	  if(err < 0) 
	    {
	    lirerr(1113);
	    }
          else
            {
            lirerr(1116);
            }
	  }
        }
      }
    else
#endif
      {
      read(rx_audio_in, rxin_isho, rx_read_bytes);
      }  
    }
  else
    {
    rxin_isho2=(void*)(&timf1_char[timf1p_pa]);
    read(rx_audio_in, rxin_isho, rx_read_bytes>>1);
    read(rx_audio_in2, rxin_isho2, rx_read_bytes>>1);
    if( (ui.rx_input_mode&DWORD_INPUT) == 0)
      {
      for(i=ad_read_fragments-1; i>=0; i--)
        {
        rxin_isho[4*i  ]=rxin_isho[2*i  ];
        rxin_isho[4*i+1]=rxin_isho[2*i+1];
        rxin_isho[4*i+2]=rxin_isho2[2*i  ];
        rxin_isho[4*i+3]=rxin_isho2[2*i+1];
        }
      }
    else
      {
      rxin_int=(void*)rxin_isho;
      rxin_int2=(void*)rxin_isho2;
      for(i=ad_read_fragments-1; i>=0; i--)
        {
        rxin_int[4*i  ]=rxin_int[2*i  ];
        rxin_int[4*i+1]=rxin_int[2*i+1];
        rxin_int[4*i+2]=rxin_int2[2*i  ];
        rxin_int[4*i+3]=rxin_int2[2*i+1];
        }
      }
    }
  finish_rx_read(rxin_isho);
  if(kill_all_flag) goto rxadin_error_exit;
  }
rxadin_error_exit:;
while( rx_audio_out == rx_audio_in ||
       tx_audio_out == rx_audio_in)lir_sleep(10000);
close_rx_sndin();
rx_adin_init_error:;
thread_status_flag[THREAD_RX_ADINPUT]=THRFLAG_RETURNED;
while(thread_command_flag[THREAD_RX_ADINPUT] != THRFLAG_NOT_ACTIVE)
  {
  lir_sleep(1000);
  }
}

void thread_test_rxda(void)
{
int i;
genparm[DA_OUTPUT_SPEED]=ui.rx_min_da_speed;
if(genparm[DA_OUTPUT_SPEED] < 8000)genparm[DA_OUTPUT_SPEED]=8000;
rx_daout_bytes=ui.rx_max_da_bytes;
rx_daout_channels=ui.rx_max_da_channels;
rx_daout_bytes_per_sample=rx_daout_bytes*rx_daout_channels;
min_delay_time=0.025;
open_rx_sndout();
if(kill_all_flag)return;
for(i=0; i<rx_daout_block; i++)
  {
  rx_da_wrbuf[i]=0; 
  }
while(!kill_all_flag &&
      soundcard_test_cmd_flag[RXDA] == THRFLAG_ACTIVE)
  {
// Write data to the output device.
  lir_rx_dawrite();
  soundcard_test_block_count[RXDA]++;
  }
close_rx_sndout();
}

void thread_test_txda(void)
{
char *buf;
int i;
tx_write_bytes=ui.tx_da_speed*ui.tx_da_bytes*ui.tx_da_channels/1000;
make_power_of_two(&tx_write_bytes);
open_tx_output();
if(kill_all_flag)return;
buf=malloc(tx_write_bytes);
if(buf == NULL)
  {
  lirerr(1245231);
  return;
  }
for(i=0; i<tx_write_bytes; i++)
  {
  buf[i]=0; 
  }
while(!kill_all_flag &&
      soundcard_test_cmd_flag[TXDA] == THRFLAG_ACTIVE)
  {
// Write data to the output device.
  lir_tx_dawrite(buf);
  soundcard_test_block_count[TXDA]++;
  }
close_tx_output();
}

void thread_test_rxad(void)
{
char *buf;
int err;
ad_read_fragments=ui.rx_ad_speed/1000;
if(ad_read_fragments < 4)ad_read_fragments=4;
make_power_of_two(&ad_read_fragments);
rx_read_bytes=2*ui.rx_ad_channels*ad_read_fragments;
if( (ui.rx_input_mode&DWORD_INPUT) != 0)rx_read_bytes*=2;
open_rx_sndin();
if(kill_all_flag)return;
buf=malloc(rx_read_bytes);
if(buf == NULL)
  {
  lirerr(1233231);
  return;
  }
while(!kill_all_flag &&
      soundcard_test_cmd_flag[RXAD] == THRFLAG_ACTIVE)
  {
  if(ui.rx_addev_no < 256)
    {
#ifdef ALSA_PRESENT
    if(ui.use_alsa)
      {
      err=snd_pcm_readi(rx_ad_handle, buf, rx_ad_frames);
      if((unsigned int)err != rx_ad_frames)
	{
        if(err == -EPIPE) //         EPIPE means overrun 
	  { 
	  err=snd_pcm_prepare(rx_ad_handle);
          if(err < 0)lirerr(1108);  
	  }
        else 
          {
	  if(err < 0) 
	    {
	    lirerr(1113);
	    }
          else
            {
            lirerr(1116);
            }
	  }
        }
      }
    else
#endif
      {
      read(rx_audio_in, buf, rx_read_bytes);
      }  
    }
  else
    {
    read(rx_audio_in, buf, rx_read_bytes>>1);
    read(rx_audio_in2, buf, rx_read_bytes>>1);
    }
  soundcard_test_block_count[RXAD]++;
  }
while( rx_audio_out == rx_audio_in ||
       tx_audio_out == rx_audio_in)lir_sleep(10000);
close_rx_sndin();
}

int lir_tx_output_samples(void)
{
int err;
#ifdef ALSA_PRESENT
if(ui.use_alsa)
  {
  if((err=alsa_getospace_txda()) != 0)
    {
    lirerr(238321);
    return -1;
    }
  }
else
#endif
  {  
  if(ioctl(tx_audio_out,SNDCTL_DSP_GETOSPACE, &tx_da_info) == -1)
                                                          lirerr(872313);
  }
return tx_output_totsamp-tx_da_info.bytes/tx_output_fragsize;
}

int lir_tx_input_samples(void)
{
#ifdef ALSA_PRESENT
if(ui.use_alsa)
  {
  return 0;
  }
else
#endif
  {
  if(ioctl(tx_audio_in,SNDCTL_DSP_GETISPACE, &tx_ad_info) == -1)lirerr(872314);
  return tx_ad_info.bytes/tx_input_fragsize;
  }
}

int lir_rx_output_bytes(void)
{
int err;
#ifdef ALSA_PRESENT
if(ui.use_alsa)
  {
  if((err=alsa_getospace_rxda())<0)return -1;
  }
else
#endif
  {  
  if(ioctl(rx_audio_out,SNDCTL_DSP_GETOSPACE, &rx_da_info) == -1)return -1;
  }
return rx_da_info.bytes;
}

int make_da_wts(void)
{
int i, err;
// Returns the number of samples or frames ( not bytes !) written to device driver but not yet sent to the speaker.
i=0;
buftest:;
#ifdef ALSA_PRESENT
if(ui.use_alsa)
  {
  if((err=alsa_getospace_rxda())<0)
    {
    rx_da_info.bytes=0;
    if(kill_all_flag) return 0;
    }
  }
else
#endif  
  {
  if(ioctl(rx_audio_out,SNDCTL_DSP_GETOSPACE, &rx_da_info) == -1)
    {
    rx_da_info.bytes=0;
    if(kill_all_flag) return 0;
    }
  }  
//************************************************************ <alsamod 
// The number of bytes we can write is incorrectly zero sometimes
// wait a little and see if we get a correct value to return.
if(rx_da_info.bytes == 0)
  {
  lir_sleep(3000);
  i++;
  if(i<5)goto buftest;
  }
i=rx_da_totbytes-rx_da_info.bytes;
i/=rx_daout_bytes_per_sample;
return i;
}

void err_restart_rxda(int errcod)
{
int i, err;
lir_sched_yield();
if(kill_all_flag || rx_audio_out == -1) return;
close_rx_sndout();
open_rx_sndout();
if(kill_all_flag) return;
for(i=0; i<rx_daout_block; i++)rx_da_wrbuf[i]=0;
#ifdef ALSA_PRESENT
if(ui.use_alsa)
  {
  err=snd_pcm_writei(rx_da_handle, rx_da_wrbuf,rx_da_frames);
  if((err=alsa_getospace_rxda())<0)   lirerr(errcod);
  }
else
#endif
  {
  write(rx_audio_out,rx_da_wrbuf,rx_daout_block);
  if(ioctl(rx_audio_out,SNDCTL_DSP_GETOSPACE,&rx_da_info) == -1)lirerr(errcod);
  }
}

void lir_empty_da_device_buffer(void)
{
int i, err;
if(rx_audio_out == -1)return;
for(i=0; i<rx_daout_block; i++)rx_da_wrbuf[i]=0;
if(rx_audio_in == rx_audio_out && 
                   rx_input_thread == THREAD_RX_ADINPUT)return;
if(tx_audio_in == rx_audio_out)return;
  {
#ifdef ALSA_PRESENT
if(ui.use_alsa)
 {
 if((err=alsa_getospace_rxda())<0)
   {
   lirerr(1077);
   return;
   }
 if(rx_da_info.fragstotal - rx_da_info.fragments <= 4)
   {
   err=snd_pcm_writei(rx_da_handle, rx_da_wrbuf,rx_da_frames) ;
   if((unsigned int)err != rx_da_frames)
    {	
    if(err == -EPIPE) // EPIPE means underrun  
	{
	err=snd_pcm_prepare(rx_da_handle);
	if(err < 0)lirerr(1117); 
	}
      else 
        {
// ESTRPIPE suspend event occurred.
// (stream is suspended and waiting for an application recovery)
        if(err == -ESTRPIPE) 
	   {
	   while((err=snd_pcm_resume(rx_da_handle)) == -EAGAIN)lir_sleep(100000);   // wait until the suspend flag is released 
 	   if(err < 0)
	     {
	     err=snd_pcm_prepare(rx_da_handle);
             if(err < 0)lirerr(1126);
             }
	   }
         else
           {
           lirerr(1127);
           }
         }
        }
      }
    }
  else
#endif  
    {
    if(ioctl(rx_audio_out,SNDCTL_DSP_GETOSPACE, &rx_da_info) == -1)
      {
      lirerr(1077);
      return;
      }
  if(rx_da_info.fragstotal - rx_da_info.fragments <= 4)
    {
    write(rx_audio_out,rx_da_wrbuf,rx_daout_block);
    }
  if(ioctl(rx_audio_out,SNDCTL_DSP_RESET,0)==-1)
    {
    lirerr(1205);
    return;
    }
   }
  }
}

void lir_tx_dawrite(char *buf)
{
int i, err;
#ifdef ALSA_PRESENT
if(ui.use_alsa)
  {
  i=0;
  while(i < tx_write_bytes)
    {
    err=snd_pcm_writei(tx_da_handle, &buf[i],tx_da_frames);
    if((unsigned int)err != tx_da_frames)
      {	
      if(err == -EPIPE) // EPIPE means under-run here 
        {
        err=snd_pcm_prepare(tx_da_handle);
        if(err < 0) 
          {
          lirerr(1140);
          }
        }
      else 
        {
// ESTRPIPE   suspend event occurred 
// (stream is suspended and waiting for an application recovery)      
        if(err == -ESTRPIPE) 
          {
// wait until the suspend flag is released         
          while ((err=snd_pcm_resume(tx_da_handle)) == -EAGAIN)
            {
            lir_sleep(100000);
            }
          if(err < 0)
            {
            err=snd_pcm_prepare(tx_da_handle);
            if(err < 0)lirerr(1144);
            }
          }
        else
          {
          if(err < 0 ) 
            { 
            lirerr(1145);
            }
          else
            {
            lirerr(1146);          
            }
          return;
          }
        }
      }
    i+=tx_daout_block;
    }
  }
else
#endif  
  {
  write(tx_audio_out,buf,tx_write_bytes);
  }
}

void lir_rx_dawrite(void)
{
int i, err;
float t1;
i=0;
buftest:;
#ifdef ALSA_PRESENT
if(ui.use_alsa)
  {
  if((err=alsa_getospace_rxda())<0)
    {
    err_restart_rxda(1109);
    if(kill_all_flag) return;
    }
  }
else
#endif  
  {
  if(ioctl(rx_audio_out,SNDCTL_DSP_GETOSPACE, &rx_da_info) == -1)
    {
    err_restart_rxda(1109);
    if(kill_all_flag) return;
    }
  }
if(rx_da_info.bytes > rx_da_maxbytes)rx_da_maxbytes=rx_da_info.bytes;
if(rx_da_info.bytes < (rx_da_totbytes>>3))
  {
// Wait for a time corresponding to 0.1 times the duration of
// the data currently in the device buffer.
  t1=rx_da_totbytes-rx_da_info.bytes;
  t1/=rx_daout_bytes_per_sample;
  t1/=genparm[DA_OUTPUT_SPEED];
  lir_sleep(t1*100000);
  i++;
  if(i<5)goto buftest;
// If we had to wait for half the time corresponding to the
// buffer content, something is very wrong.
// Force a restart...
  err_restart_rxda(23855);
  if(kill_all_flag) return;
  }
#ifdef ALSA_PRESENT
if(ui.use_alsa)
  {
  err=snd_pcm_writei(rx_da_handle, rx_da_wrbuf,rx_da_frames);
  if((unsigned int)err != rx_da_frames)
    {	
    if(err == -EPIPE) // EPIPE means under-run here 
      {
      err=snd_pcm_prepare(rx_da_handle);
      if(err < 0) 
        {
        lirerr(1140);
        }
      }
    else 
      {
// ESTRPIPE   suspend event occurred 
// (stream is suspended and waiting for an application recovery)      
      if(err == -ESTRPIPE) 
        {
// wait until the suspend flag is released         
        while ((err=snd_pcm_resume(rx_da_handle)) == -EAGAIN) lir_sleep(100000);
        if(err < 0)
          {
          err=snd_pcm_prepare(rx_da_handle);
          if(err < 0)lirerr(1144);
          }
        }
      else
        {
        if(err < 0 ) 
          { 
          lirerr(1145);
          }
        else
          {
          lirerr(1146);          
          }
        }
      }
    }
  }
else
#endif  
  {
  write(rx_audio_out,rx_da_wrbuf,rx_daout_block);
  }
}

void sndlog_sync(void)
{
fflush(sndlog);
sync();
}

void make_devname(int nn, char* ss)
{
int n;
n=nn&255;
// This routine is greatly simplified July 2006 thanks to Diane Bruce, VA3DB.
if(n>MAX_DEVNAMES)
  {
  lirerr(1255);
  return;
  }
if(n+2 == MAX_DEVNAMES)
  {
  sprintf(ss, "%s", DEVNAME_BASE);
  }
else
  {
  sprintf(ss, "%s%d", DEVNAME_BASE, n);
  }
}

void close_rx_sndin(void)
{
int err;
if(rx_audio_in != -1)
  {
  if(rx_audio_out == rx_audio_in)
    {
    close_rx_sndout();
    }
#ifdef ALSA_PRESENT
  if(ui.use_alsa)
    {
    err=snd_pcm_close(rx_ad_handle);
    if(err < 0) 
      {
      lirerr(1005);
      return;
      }  
     }
  else
#endif
    {
    if(close(rx_audio_in) == -1)
      {
      lirerr(1005);
      return;
      }
    }
  if(ui.rx_addev_no > 255)
    {
    if(close(rx_audio_in2) == -1)lirerr(1173);
    }
  }
rx_audio_in=-1;
}

void open_tx_input(void)
{
unsigned int val;
int ad_bytes_per_sample;
static snd_pcm_format_t format;
snd_pcm_hw_params_t *hw_ad_params;
char ss[sizeof(DEVNAME_BASE)+2];
int i, j, err, frag, resample;
float t1;
if(ui.tx_addev_no == ui.rx_dadev_no && ui.rx_damode == O_RDWR)
  {
  while(rx_audio_out == -1)
    {
    lir_sleep(10000);
    }
  tx_audio_in=rx_audio_out; 
  return;
  }
#ifdef ALSA_PRESENT
if(ui.use_alsa)
  {
  err=alsa_get_dev_id (ui.tx_addev_no);
  if(err<0)
    {
    lirerr(1255);
    return;
    }
  sprintf(sndin_plughw_pcm_name,"%s", alsa_dev_plughw_pcm_name);
//**********************************************************************
//********************* OPEN ALSA CAPTURE DEVICE   *********************
//**********************************************************************
  err=snd_pcm_open(&tx_ad_handle, sndin_plughw_pcm_name, 
                                                     SND_PCM_STREAM_CAPTURE,0);
  if(err < 0) 
    {
    lirerr(1271);
    return;
    }
// use tx_audio_in as switch for compatibility reasons with the existing OSS logic */
  tx_audio_in=(int)tx_ad_handle;
//alsa housekeeping
  err=snd_pcm_hw_params_malloc(&hw_ad_params);
  if(err < 0) 
    {
    lirerr(1272);
    return;
    }
  err=snd_pcm_hw_params_any(tx_ad_handle, hw_ad_params);
  if(err < 0)
    {
    lirerr(1282);
    return;
    }
// Interleaved mode 
  err=snd_pcm_hw_params_set_access(tx_ad_handle, hw_ad_params,
                                             SND_PCM_ACCESS_RW_INTERLEAVED);
  if(err< 0) 
    {
    lirerr(1294);
    return;
    }
// Set sample format 
  if(ui.tx_ad_bytes == 2)
    {
    format=SND_PCM_FORMAT_S16_LE;
    }
  else
    {  
    format=SND_PCM_FORMAT_S32_LE;
    }
  ad_bytes_per_sample=snd_pcm_format_width(format)/8;
  err=snd_pcm_hw_params_set_format(tx_ad_handle, hw_ad_params,format);
  if(err < 0)
    {
    lirerr(1295);
    return;
    }
// Set number of ad_channels 
  err=snd_pcm_hw_params_set_channels(tx_ad_handle, 
                                           hw_ad_params, ui.tx_ad_channels);
  if(err < 0) 
    {
    lirerr(1296);
    return;
    }
// Set sample rate  
// first enable  hardware resample 
  resample=1;
  err=snd_pcm_hw_params_set_rate_resample(tx_ad_handle,hw_ad_params, resample);
  if(err < 0) 
    {
    lirerr(1298);
    return;
    } 
// now set sample rate  
  val=ui.tx_ad_speed;
  err=snd_pcm_hw_params_set_rate_near(tx_ad_handle,hw_ad_params, &val, 0);
  if(err < 0) 
    {
    lirerr(1299);
    return;
    } 
  t1=(float)(val)/ui.tx_ad_speed;
  if(t1 > 1.05 || t1 < 0.95)
    {
    lirerr(1012);
    return;
    }
  ui.tx_ad_speed=(int)val;
// Set period size:
// We want the period_size to match tx_read_fragments 
  tx_read_fragments=tx_read_bytes/(ui.tx_ad_bytes*ui.tx_ad_channels);
  val=tx_read_fragments;
periodset:;
  err=snd_pcm_hw_params_set_period_size(tx_ad_handle,hw_ad_params, val, 0);
  if(err < 0)
    {
    val/=2;
    if(val > 256)goto periodset;
    lirerr(1300);
    return;
    }
//  Write all the parameters to the driver 
  err=snd_pcm_hw_params(tx_ad_handle, hw_ad_params);
  if(err < 0) 
    {
    lirerr(1301);
    return;
    }
// free the memory for the hardware parameter structure 
  snd_pcm_hw_params_free(hw_ad_params);	
  err=snd_pcm_start(tx_ad_handle);
  if(err < 0) 
    {
    lirerr(1302);
    }
  t1=(float)(ui.tx_ad_speed)*ad_bytes_per_sample/(val*ui.tx_ad_bytes);
  tx_ad_frames=tx_read_bytes/(ui.tx_ad_channels*ad_bytes_per_sample);
  DEB"\nTXin interrupt_rate: Desired  %f",t1);
  ioctl(tx_audio_in, SNDCTL_DSP_GETBLKSIZE,&i);
  t1=(float)(ui.tx_ad_speed)/tx_ad_frames;
  DEB"  actual %f\n",t1);
  }
else
#endif  	  
  {
  make_devname(ui.tx_addev_no, ss);
  tx_audio_in=open( ss ,O_RDONLY, 0);
  if(tx_audio_in == -1)
    {
    lirerr(1260);
    return;
    }
// We want the fragment size to match tx_read_bytes.
  frag=0;
  i=tx_read_bytes;
  while(i>1)
    {
    i>>=1;
    frag++;
    }
  if(frag < 5)frag=5;
  frag=frag|0x7fff0000;
  if(ioctl(tx_audio_in, SNDCTL_DSP_SETFRAGMENT, &frag) == -1)
    {
    lirerr(1262);
    return;
    }
  i=AFMT_S16_LE;
#ifdef AFMT_S32_LE
  if(ui.tx_ad_bytes==4)i=AFMT_S32_LE;
#else
  ui.tx_ad_bytes=2;
#endif
  j=i;
  if(ioctl(tx_audio_in, SNDCTL_DSP_SETFMT, &i) == -1)
    {
    lirerr(231013);
    return;
    }
  if(i!=j)
    {
    lirerr(231014);
    return;
    }
// Set number of channels
  j=ui.tx_ad_channels;
  i=j;
  if(ioctl(tx_audio_in, SNDCTL_DSP_CHANNELS, &i) == -1)
    {
    lirerr(231009);  
    return;
    }
  if(i != j)
    {
    lirerr(231010);
    return;
    }
  i=ui.tx_ad_speed;
  if(ioctl(tx_audio_in, SNDCTL_DSP_SPEED, &i)==-1)
    {
    lirerr(231011);
    return;
    }
  t1=(float)(i)/ui.tx_ad_speed;
  if(t1 > 1.05 || t1 < 0.95)
    {
    lirerr(231012);
    return;
    }
  ui.tx_ad_speed=i;
// This has to be the last ioctl call before read or write!!
  t1=(float)(ui.tx_ad_channels*ui.tx_ad_bytes*ui.tx_ad_speed)/tx_read_bytes;
  DEB"\nTXin interrupt_rate: Desired  %f",t1);
  ioctl(tx_audio_in, SNDCTL_DSP_GETBLKSIZE,&i);
  t1=(float)(ui.tx_ad_channels*ui.tx_ad_bytes*ui.tx_ad_speed)/i;
  DEB"  actual %f\n",t1);
  }
}

void close_tx_input(void)
{
int err;
if(ui.tx_addev_no == ui.rx_dadev_no && ui.rx_damode == O_RDWR)
  {
  tx_audio_in=-1; 
  return;
  }
#ifdef ALSA_PRESENT
if(ui.use_alsa)
  {
  err=snd_pcm_close(tx_ad_handle);
  if(err < 0) 
    {
    lirerr(1005);
    return;
    }  
   }
else
#endif
  {
  if(close(tx_audio_in)==-1)lirerr(231037);  
  tx_audio_in=-1;
  }
}

void open_tx_output(void)
{
char *buf;
unsigned int val;
static snd_pcm_format_t format;
snd_pcm_hw_params_t *hw_da_params;
int i, j, dir, err, frag, resample;
char ss[sizeof(DEVNAME_BASE)+2];
float t1;
if(ui.tx_dadev_no == ui.rx_addev_no && ui.rx_admode == O_RDWR)
  {
  while(rx_audio_in == -1)
    {
    lir_sleep(30000);
    }
  tx_audio_out=rx_audio_in; 
  goto ok_exit;
  }
#ifdef ALSA_PRESENT
//*********************************************************************
//********************  OPEN ALSA PLAYBACK DEVICE *********************
//*********************************************************************
if(ui.use_alsa)
  {
  if((err=alsa_get_dev_id (ui.tx_dadev_no))<0)
    {
    lirerr(1255);
    return;
    }
  sprintf(sndout_plughw_pcm_name,"%s", alsa_dev_plughw_pcm_name);  //multithreading !!
  if((err=snd_pcm_open(&tx_da_handle, sndout_plughw_pcm_name, 
                                         SND_PCM_STREAM_PLAYBACK, 0)) < 0) 
    {
    lirerr(1304);
    return;
    }
// use tx_audio_out as switch for compatibility reasons with the existing OSS logic 
  tx_audio_out=(int)tx_da_handle;
//*********************************************************************
//******************** SET PLAYBACK PARAMETERS THE ALSA WAY ******
//*********************************************************************
  tx_daout_block=tx_write_bytes;
  if(tx_daout_block > 256)tx_daout_block=256; 
  tx_output_blockrate=(float)(ui.tx_da_bytes*ui.tx_da_channels*
                                 ui.tx_da_speed)/tx_daout_block;
  while(tx_output_blockrate > ui.max_dma_rate && tx_daout_block < 256)
    {
    tx_output_blockrate/=2;
    tx_daout_block*=2;
    }
  err=snd_pcm_hw_params_malloc (&hw_da_params);
  if(err < 0) 
    {
    lirerr(1305);
    return;
    }
  err=snd_pcm_hw_params_any(tx_da_handle, hw_da_params);
  if(err < 0) 
    {
    lirerr(1306);
    return;
    }	
// Interleaved mode 
  err=snd_pcm_hw_params_set_access(tx_da_handle, hw_da_params,
                                         SND_PCM_ACCESS_RW_INTERLEAVED);
  if(err < 0) 
    {
    lirerr(1307);
    return;
    }
// Set sample format 
  if(ui.tx_da_bytes == 2)
    {
    format=SND_PCM_FORMAT_S16_LE;
    }
  else
    {
    format=SND_PCM_FORMAT_S32_LE;
    }
  err=snd_pcm_hw_params_set_format(tx_da_handle, hw_da_params,format);
  if(err < 0)
    {
    lirerr(1308);
    return;
    }
// Set number of da_channels 
  err=snd_pcm_hw_params_set_channels(
                               tx_da_handle, hw_da_params, ui.tx_da_channels);
  if(err <0) 
    {
    lirerr(1309);
    return;
    }
// Set sample rate  
// first enable  hardware resample 
  resample=1;
  err=snd_pcm_hw_params_set_rate_resample(
                                     tx_da_handle, hw_da_params, resample);
  if(err < 0) 
    {
    lirerr(1310);
    return;
    } 
// now set sample rate  
  val=ui.tx_da_speed;
  if((err =snd_pcm_hw_params_set_rate_near(
                                   tx_da_handle,hw_da_params, &val, 0)<0)) 
    {
    lirerr(1311);
    return;
    } 
  t1=(float)(val)/ui.tx_da_speed;
  if(t1 > 1.05 || t1 < 0.95)
    {
    lirerr(1023);
    return;
    }
  ui.tx_da_speed=val;
// set the period size ( in frames )  to  (tx_daout_block/tx_daout_bytes_per_sample) 
// note:  size of a frame is == tx_daout_bytes_per_sample == (tx_daout_bytes*tx_daout_channels) 
  tx_da_bytes_per_frame=ui.tx_da_bytes*ui.tx_da_channels;
  tx_da_frames=tx_daout_block/tx_da_bytes_per_frame;  
  DEB"\nopen_tx_output:Desired period_size for playback: %i frames",
                                                         (int)tx_da_frames);
  dir=0; 
// obtained value is <,=,> target-value, according value of  dir (-1,0,1) 
  err=snd_pcm_hw_params_set_period_size_near(tx_da_handle, hw_da_params,
                                                        &tx_da_frames, &dir);
  if(err < 0) 
    {
    lirerr(1312);
    return;
    }
// check obtained  period size 
  err=snd_pcm_hw_params_get_period_size(hw_da_params, &tx_da_period_size, &dir);
  if(err < 0) 
   {
   DEB"\nopen_tx_output:Unable to get period size for playback: %s\n", 
                                                       snd_strerror(err));
   }
  DEB"\nopen_tx_output:Obtained period size for playback: %i frames",
                                                        (int) tx_da_period_size);
// set the hw parameters 
  err=snd_pcm_hw_params(tx_da_handle, hw_da_params);
  if(err < 0) 
    {
    lirerr(1313);
    return;
    }
  err=snd_pcm_prepare(tx_da_handle);
  if(err < 0) 
    {
    lirerr(1314);
    return;
    }
// free the memory for the hardware parameter struct 
  snd_pcm_hw_params_free (hw_da_params);	
  tx_da_frames= tx_daout_block/tx_da_bytes_per_frame;
// get buffersize
  err=snd_pcm_get_params ( tx_da_handle, &tx_da_buffer_size, &tx_da_period_size );	
  if(err == -EPIPE) 
    {
// * under-run *
    err=snd_pcm_prepare(tx_da_handle);
    if(err < 0)
      {
      lirerr(1313);
      return;
      }
    }  
  buf=malloc(tx_daout_block);
  if(buf == NULL)
    {
    lirerr(1025);
    return;
    }
  for(i=0; i<tx_daout_block; i++)buf[i]=0;
  lir_tx_dawrite(buf);
  free(buf);
  alsa_getospace_txda();
  tx_da_totbytes=tx_da_info.fragstotal*tx_da_info.fragsize;
  }
else
#endif    
  {
  make_devname(ui.tx_dadev_no, ss);
  tx_audio_out=open( ss, O_WRONLY|O_NONBLOCK, 0);
  if(tx_audio_out == -1)
    {
    lirerr(1079);
    return;
    }
  frag=0;
  i=tx_write_bytes;
  while(i>1)
    {
    i>>=1;
    frag++;
    }
  if(frag < 5)frag=5;
  frag=frag|0x7fff0000;
  if(ioctl(tx_audio_out, SNDCTL_DSP_SETFRAGMENT, &frag) == -1)
    {
    lirerr(991262);
    return;
    }
  i=AFMT_S16_LE;
#ifdef AFMT_S32_LE
  if(ui.tx_da_bytes==4)i=AFMT_S32_LE;
#else
  ui.tx_da_bytes=2;
#endif
  j=i;
  if(ioctl(tx_audio_out, SNDCTL_DSP_SETFMT, &i) == -1)
    {
    lirerr(231013);
    return;
    }
  if(i!=j)
    {
    lirerr(231014);
    return;
    }
// Set number of channels
  j=ui.tx_da_channels;
  i=j;
  if(ioctl(tx_audio_out, SNDCTL_DSP_CHANNELS, &i) == -1)
    {
    lirerr(231009);  
    return;
    }
  if(i != j)
    {
    lirerr(231010);
    return;
    }
  i=ui.tx_da_speed;
  if(ioctl(tx_audio_out, SNDCTL_DSP_SPEED, &i)==-1)
    {
    lirerr(231011);
    return;
    }
  t1=(float)(i)/ui.tx_da_speed;
  if(t1 > 1.05 || t1 < 0.95)
    {
    lirerr(231112);
    return;
    }
  ui.tx_da_speed=i;
// This has to be the last ioctl call before read or write!!
  ioctl(tx_audio_in, SNDCTL_DSP_GETBLKSIZE,&i);
  }
ok_exit:;
tx_output_totsamp=0;
tx_output_totsamp=-lir_tx_output_samples();
}

void close_tx_output(void)
{
int err;
if(ui.tx_dadev_no == ui.rx_addev_no && ui.rx_admode == O_RDWR)
  {
  tx_audio_out=-1; 
  return;
  }
#ifdef ALSA_PRESENT
if(ui.use_alsa)
  {
  snd_pcm_drop(tx_da_handle);
  err= snd_pcm_close(tx_da_handle);
  if(err < 0)lirerr(1006);
  }
else
#endif    
  {
  if(ioctl(tx_audio_out,SNDCTL_DSP_RESET,0)==-1)
    {
    lirerr(1288);
    }
  if(close(tx_audio_out)==-1)lirerr(1287);  
  tx_audio_out=-1;
  }
}

int make_dafrag(void)
{
int i, frag;
// We want the fragment size to match rx_output_blockrate,
// the rate at which rx_daout_block is written to the output.
frag=0;
i=rx_daout_block>>1;
while(i>0)
  {
  i>>=1;
  frag++;
  }
if(frag < 4)frag=4;
frag|=0x7fff0000;
return frag;
}

void close_rx_sndout(void)
{
int err;
if(rx_audio_out == -1)
  {
  return;
  }
#ifdef ALSA_PRESENT
if(ui.use_alsa)
  {
  snd_pcm_drop(rx_da_handle);
  err= snd_pcm_close(rx_da_handle);
  if(err < 0)lirerr(1006);
  }
else
#endif    
  {
  if(rx_audio_out != rx_audio_in)
    {
    if(ioctl(rx_audio_out,SNDCTL_DSP_RESET,0)==-1)
      {
      lirerr(1026);
      }
    if(close(rx_audio_out)==-1)lirerr(1006);
    }
  }
if(rx_da_wrbuf != NULL)
  {
  free(rx_da_wrbuf);
  rx_da_wrbuf=NULL;
  }
rx_audio_out=-1;
}

void open_rx_sndin(void)
{
int ad_bytes_per_sample;
unsigned int val;
static snd_pcm_format_t format;
snd_pcm_hw_params_t *hw_ad_params;
char ss[sizeof(DEVNAME_BASE)+2];
float t1;
int i, j, err, frag, resample;
audio_buf_info ad_info;
if( (rx_audio_in) != -1)
  {
  if(ui.rx_admode == O_RDWR)return;
  lirerr(1000);
  return;
  }
#ifdef ALSA_PRESENT
if(ui.use_alsa)
  {
  if((err=alsa_get_dev_id (ui.rx_addev_no))<0)
    {
    lirerr(1255);
    return;
    }
  sprintf(sndin_plughw_pcm_name,"%s", alsa_dev_plughw_pcm_name);  //multithreading !!
//**********************************************************************
//********************* OPEN ALSA CAPTURE DEVICE   *********************
//**********************************************************************
  if((err=snd_pcm_open(&rx_ad_handle, sndin_plughw_pcm_name, 
                                         SND_PCM_STREAM_CAPTURE, 0)) < 0) 
    {
    lirerr(1271);
    return;
    }
// use rx_audio_in as switch for compatibility reasons with the existing OSS logic */
  rx_audio_in=(int)rx_ad_handle;
//alsa housekeeping
  if((err=snd_pcm_hw_params_malloc (&hw_ad_params)) < 0) 
    {
    lirerr(1272);
    return;
    }
  if((err=snd_pcm_hw_params_any(rx_ad_handle, hw_ad_params)) < 0)
    {
    lirerr(1282);
    return;
    }
// Interleaved mode 
  if((err=snd_pcm_hw_params_set_access(rx_ad_handle, hw_ad_params,
                                  SND_PCM_ACCESS_RW_INTERLEAVED))< 0) 
    {
    lirerr(1294);
    return;
    }
// Set sample format 
  format=SND_PCM_FORMAT_S16_LE;
  if( (ui.rx_input_mode&DWORD_INPUT) != 0) format=SND_PCM_FORMAT_S32_LE;
  ad_bytes_per_sample=snd_pcm_format_width(format)/8;
  err=snd_pcm_hw_params_set_format (rx_ad_handle, hw_ad_params,format);
  if(err < 0)
    {
    lirerr(1295);
    return;
    }
// Set number of ad_channels 
  err=snd_pcm_hw_params_set_channels(rx_ad_handle, 
                                           hw_ad_params, ui.rx_ad_channels);
  if(err < 0) 
    {
    lirerr(1296);
    return;
    }
// Set sample rate  
// first enable  hardware resample 
  resample=1;
  err=snd_pcm_hw_params_set_rate_resample(rx_ad_handle,hw_ad_params,resample);
  if(err < 0)
    {
    lirerr(1298);
    return;
    } 
// now set sample rate  
  val=ui.rx_ad_speed;
  err=snd_pcm_hw_params_set_rate_near(rx_ad_handle,hw_ad_params, &val, 0);
  if(err<0) 
    {
    lirerr(1299);
    return;
    } 
  t1=(float)(val)/ui.rx_ad_speed;
  if(t1 > 1.05 || t1 < 0.95)
    {
    lirerr(1012);
    return;
    }
  ui.rx_ad_speed=(int)val;
// Set period size:
// We want the period_size to match ad_read_fragments 
  val=ad_read_fragments;
periodset:;
  err=snd_pcm_hw_params_set_period_size(rx_ad_handle,hw_ad_params, val, 0);
  if(err<0)
    {
    val/=2;
    if(val > 256)goto periodset;
    lirerr(1300);
    return;
    }
//  Write all the parameters to the driver 
  if((err=snd_pcm_hw_params(rx_ad_handle, hw_ad_params))< 0) 
    {
    lirerr(1301);
    return;
    }
// free the memory for the hardware parameter structure 
  snd_pcm_hw_params_free (hw_ad_params);	
  err=snd_pcm_start(rx_ad_handle);
  if(err < 0) 
    {
    lirerr(1302);
    return;
    }
  rx_ad_frames=rx_read_bytes/(ui.rx_ad_channels*ad_bytes_per_sample);
  return;
  }
else
#endif  	  
  {
  make_devname(ui.rx_addev_no, ss);
  rx_audio_in=open( ss ,ui.rx_admode , 0);
  if(rx_audio_in == -1)
    {
    lirerr(1007);
    return;
    }
  if(ui.rx_addev_no > 255)
    {
    i=ui.rx_addev_no/255-1;
    make_devname(i, ss);
    rx_audio_in2=open( ss ,ui.rx_admode , 0);
    if(rx_audio_in2 == -1)
      {
      lirerr(1174);
      return;
      }
    }
// We want the fragment size to match rx_read_bytes.
  frag=0;
  i=rx_read_bytes>>1;
  while(i>0)
    {
    i>>=1;
    frag++;
    }
  if(frag < 4)frag=4;
  if(ui.rx_addev_no > 255)
    {
    frag--;
    if(frag < 4)frag=4;
    frag=frag|0x7fff0000;
    if(ioctl(rx_audio_in2, SNDCTL_DSP_SETFRAGMENT, &frag) == -1)
      {
      lirerr(1175);
      return;
      }
    }
  frag=frag|0x7fff0000;
  if(ioctl(rx_audio_in, SNDCTL_DSP_SETFRAGMENT, &frag) == -1)
    {
    lirerr(1008);
    return;
    }
  i=AFMT_S16_LE;
#ifdef AFMT_S32_LE
  if( (ui.rx_input_mode&DWORD_INPUT) != 0)i=AFMT_S32_LE;
#else
  if( (ui.rx_input_mode&DWORD_INPUT) != 0)
    {
    lirerr(1246);
    return;
    }
#endif
  j=i;
  if(ioctl(rx_audio_in, SNDCTL_DSP_SETFMT, &i) == -1)
    {
    lirerr(1013);
    return;
    }
  if(i!=j)
    {
    lirerr(1014);
    return;
    }
  if(ui.rx_addev_no > 255)
    {
    j=i;
    if(ioctl(rx_audio_in2, SNDCTL_DSP_SETFMT, &i) == -1)
      {
      lirerr(1176);
      return;
      }
    if(i!=j)
      {
      lirerr(1177);
      return;
      }
    }
// Set number of ad_channels
  j=ui.rx_ad_channels;
  if(ui.rx_addev_no > 255)j=2;
  i=j;
  if(ioctl(rx_audio_in, SNDCTL_DSP_CHANNELS, &i) == -1)
    {
    lirerr(1009);
    return;
    }
  if(i != j)
    {
    lirerr(1010);
    return;
    }
  if(ui.rx_addev_no > 255)
    {
    i=j;
    if(ioctl(rx_audio_in2, SNDCTL_DSP_CHANNELS, &i) == -1)
      {
      lirerr(1178);
      return;
      }
    if(i != j)
      {
      lirerr(1179);
      return;
      }
    }
  i=ui.rx_ad_speed;
  if(ioctl(rx_audio_in, SNDCTL_DSP_SPEED, &i)==-1)
    {
    lirerr(1011);
    return;
    }
  t1=(float)(i)/ui.rx_ad_speed;
  if(t1 > 1.05 || t1 < 0.95)
    {
    lirerr(1012);
    return;
    }
  ui.rx_ad_speed=i;
  if(ui.rx_addev_no > 255)
    {
    if(ioctl(rx_audio_in2, SNDCTL_DSP_SPEED, &i)==-1)
      {
      lirerr(1180);
      return;
      }
    if(i != ui.rx_ad_speed)
      {
      lirerr(1181);
      return;
      }
// Now that both devices are configured to be equal, synchronize them.
    i=0;
    if(ioctl(rx_audio_in,SNDCTL_DSP_SETTRIGGER,&i)==-1)
      {
      lirerr(1183);
      return;
      }
    i=0;
    if(ioctl(rx_audio_in2,SNDCTL_DSP_SETTRIGGER,&i)==-1)
      {
      lirerr(1184);
      return;
      }
    i=PCM_ENABLE_INPUT;
    if(ioctl(rx_audio_in,SNDCTL_DSP_SETTRIGGER,&i)==-1)
      {
      lirerr(1183);
      return;
      }
    i=PCM_ENABLE_INPUT;
    if(ioctl(rx_audio_in2,SNDCTL_DSP_SETTRIGGER,&i)==-1)
      {
      lirerr(1184);
      return;
      }
    }
  if(ioctl(rx_audio_in, SNDCTL_DSP_GETISPACE, &ad_info) == -1)
    {
    lirerr(1109);
    return;
    }
  overrun_limit=ad_info.fragstotal-1;
  if(ui.rx_addev_no > 255)
    {
    if(ioctl(rx_audio_in2, SNDCTL_DSP_GETISPACE, &ad_info) == -1)
      {
      lirerr(1109);
      return;
      }
    if(overrun_limit != ad_info.fragstotal-1)
      {
      lirerr(1109);
      }
    }
// This has to be the last ioctl call before read or write!!
  ioctl(rx_audio_in,SNDCTL_DSP_GETBLKSIZE,&i);
  }
}

void open_rx_sndout(void)
{
unsigned int val;
static snd_pcm_format_t format;
snd_pcm_hw_params_t *hw_da_params;
char ss[sizeof(DEVNAME_BASE)+2];
int i, j, dir, err, resample;
int old_disksave_flag, old_mode, old_chan, old_speed;
float t1;
int frag;
if( rx_audio_out != -1)
  {
  lirerr(1015);
  return;
  }
if( (ui.rx_input_mode&NO_DUPLEX) != 0 && rx_audio_in != -1)return;
if(ui.rx_dadev_no == -1)
  {
  if(disksave_flag >=2)
    {
    if(rx_audio_in != -1)
      {
      lirerr(9865);
      return;
      }
#ifdef ALSA_PRESENT
//*********************************************************************
//********************  OPEN ALSA PLAYBACK DEVICE *********************
//*********************************************************************
    if(ui.use_alsa)
      {
      if((err=alsa_get_dev_id (ui.rx_dadev_no))<0)
        {
        lirerr(1255);
        return;
        }
      sprintf(sndout_plughw_pcm_name,"%s", alsa_dev_plughw_pcm_name);  //multithreading !!
      if((err=snd_pcm_open(&rx_da_handle, sndout_plughw_pcm_name, 
                                            SND_PCM_STREAM_PLAYBACK, 0)) < 0) 
        {
        lirerr(1303);
        return;
        }
// use rx_audio_out as switch for compatibility with the existing OSS logic 
      rx_audio_out=(int)rx_da_handle;
      }
    else
#endif
      {  
      make_devname(ui.rx_addev_no, ss);
      rx_audio_out=open( ss ,O_WRONLY , 0);
      }
    if(rx_audio_out == -1)
      {
      lirerr(1086);
      return;
      }
    goto set_da;
    }
  else
    {
    return;
    }
  }
// *******************************************
if(ui.rx_addev_no == ui.rx_dadev_no &&
   ui.rx_admode == O_RDWR )
  {
  old_disksave_flag=disksave_flag;
  disksave_flag=0;
  old_mode=ui.rx_input_mode;
  ui.rx_input_mode=0;
  old_chan=ui.rx_ad_channels;
  ui.rx_ad_channels=rx_daout_channels;
  old_speed=ui.rx_ad_speed;
  ui.rx_ad_speed=genparm[DA_OUTPUT_SPEED];
  if(rx_audio_in == -1)
    {
    open_rx_sndin();
    if(kill_all_flag) return;
    }
  rx_audio_out=rx_audio_in;
  disksave_flag=old_disksave_flag;
  ui.rx_input_mode=old_mode;
  ui.rx_ad_channels=old_chan;
  ui.rx_ad_speed=old_speed;
  }
else
  {
#ifdef ALSA_PRESENT
//*********************************************************************
//********************  OPEN ALSA PLAYBACK DEVICE *********************
//*********************************************************************
  if(ui.use_alsa)
    {
    err=alsa_get_dev_id (ui.rx_dadev_no);
    if(err < 0)
      {
      lirerr(1255);
      return;
      }
    sprintf(sndout_plughw_pcm_name,"%s", alsa_dev_plughw_pcm_name);  //multithreading !!
    err=snd_pcm_open(&rx_da_handle, sndout_plughw_pcm_name, 
                                                 SND_PCM_STREAM_PLAYBACK, 0);
    if(err < 0) 
      {
      lirerr(1304);
      return;
      }
// use rx_audio_out as switch for compatibility reasons with the existing OSS logic 
    rx_audio_out=(int)rx_da_handle;
    }
  else
#endif    
    {
    make_devname(ui.rx_dadev_no, ss);
    rx_audio_out=open( ss ,ui.rx_damode , 0);
    }
  if(rx_audio_out == -1)
    {
    if(disksave_flag==2)
      {
      lirerr(1191);
      }
    else
      {
      lirerr(1017);
      }
    return;
    }
set_da:;
  t1=min_delay_time;
  i=4;
  if(disksave_flag >= 2)i=1;
  if(t1 < fft3_blocktime/i)t1=fft3_blocktime/i;
  if(t1 > 1)t1=1;
  rx_daout_block=0.75*rx_daout_bytes_per_sample*t1*genparm[DA_OUTPUT_SPEED];
  make_power_of_two(&rx_daout_block);
// make the minimum block 32 bytes in accordance with the minimum
// fragment size of OSS (frag=4)
  if(rx_daout_block < 32)rx_daout_block=32;
#ifdef ALSA_PRESENT
  if(ui.use_alsa)
    {
//*********************************************************************
//******************** SET PLAYBACK PARAMETERS THE ALSA WAY ******
//*********************************************************************
    if(rx_daout_block > 256)rx_daout_block=256; 
    rx_output_blockrate=(float)(rx_daout_bytes_per_sample*
                                 genparm[DA_OUTPUT_SPEED])/rx_daout_block;
    while(rx_output_blockrate > ui.max_dma_rate && rx_daout_block < 256)
      {
      rx_output_blockrate/=2;
      rx_daout_block*=2;
      }
    err=snd_pcm_hw_params_malloc (&hw_da_params);
    if(err < 0) 
      {
      lirerr(1305);
      return;
      }
    err=snd_pcm_hw_params_any(rx_da_handle, hw_da_params);
    if(err < 0) 
      {
      lirerr(1306);
      return;
      }	
// Interleaved mode 
    err=snd_pcm_hw_params_set_access(rx_da_handle, hw_da_params,
                                           SND_PCM_ACCESS_RW_INTERLEAVED);
    if(err < 0) 
      {
      lirerr(1307);
      return;
      }
// Set sample format 
    if(rx_daout_bytes == 1)
      {
      format=SND_PCM_FORMAT_U8;
      }
    else
      {
      format=SND_PCM_FORMAT_S16_LE;
      }
    err=snd_pcm_hw_params_set_format(rx_da_handle, hw_da_params,format);
    if(err < 0)
      {
      lirerr(1308);
      return;
      }
// Set number of da_channels
    err=snd_pcm_hw_params_set_channels(
                              rx_da_handle, hw_da_params, rx_daout_channels);
    if(err <0 ) 
      {
      lirerr(1309);
      return;
      }
// Set sample rate  
// first enable  hardware resample 
    resample=1;
    err=snd_pcm_hw_params_set_rate_resample(
                                  rx_da_handle,hw_da_params, resample);
    if(err < 0) 
      {
      lirerr(1310);
      return;
      } 
// now set sample rate  
    val=genparm[DA_OUTPUT_SPEED];
    err=snd_pcm_hw_params_set_rate_near(rx_da_handle,hw_da_params, &val, 0);
    if(err < 0) 
      {
      lirerr(1311);
      return;
      } 
    t1=(float)(val)/genparm[DA_OUTPUT_SPEED];
    if(t1 > 1.05 || t1 < 0.95)
      {
      lirerr(1023);
      return;
      }
    genparm[DA_OUTPUT_SPEED]=val;
// set the period size ( in frames )  to  (rx_daout_block/rx_daout_bytes_per_sample) 
// note:  size of a frame is == rx_daout_bytes_per_sample == (rx_daout_bytes*rx_daout_channels) 
    DEB"\nopen_rx_sndout:Desired period_size for playback: %i frames",
                                  (rx_daout_block/rx_daout_bytes_per_sample));

    rx_da_frames=(rx_daout_block/rx_daout_bytes_per_sample);
    dir=0; 
// obtained value is <,=,> target-value, according value of  dir (-1,0,1) 
    err=snd_pcm_hw_params_set_period_size_near(rx_da_handle, hw_da_params,
                                                        &rx_da_frames, &dir);
    if(err < 0) 
      {
      lirerr(1312);
      return;
      }
// check obtained  period size 
    err=snd_pcm_hw_params_get_period_size(hw_da_params, &rx_da_period_size, &dir);
    if(err < 0) 
     {
     DEB"\nopen_rx_sndout:Unable to get period size for playback: %s\n", 
                                                       snd_strerror(err));
     }
    DEB"\nopen_rx_sndout:Obtained period size for playback: %i frames",
                                                        (int) rx_da_period_size);
    DEB"\nRXout interrupt_rate:");
// set the hw parameters 
    err=snd_pcm_hw_params(rx_da_handle, hw_da_params);
    if(err < 0) 
      {
      lirerr(1313);
      return;
      }
    err=snd_pcm_prepare(rx_da_handle);  
    if(err < 0) 
      {
      lirerr(1314);
      return;
      }
// free the memory for the hardware parameter struct 
    snd_pcm_hw_params_free(hw_da_params);	
    rx_da_bytes_per_frame=rx_daout_bytes_per_sample;
    rx_da_frames=rx_daout_block/rx_da_bytes_per_frame;
// get buffersize
    err=snd_pcm_get_params ( rx_da_handle, &rx_da_buffer_size, &rx_da_period_size );	
    if(err == -EPIPE) 
      {
// * under-run *
      err=snd_pcm_prepare(rx_da_handle);
      if(err < 0)
        {
        lirerr(1313);
        return;
        }
      }  
    rx_da_wrbuf=malloc(rx_daout_block);
    if(rx_da_wrbuf == NULL)
      {
      lirerr(1025);
      return;
      }
    for(i=0; i<rx_daout_block; i++)rx_da_wrbuf[i]=0;
    lir_rx_dawrite();
    alsa_getospace_rxda();
    rx_da_totbytes=rx_da_info.fragstotal*rx_da_info.fragsize;
    }
  else
#endif
    {
    rx_output_blockrate=(float)(rx_daout_bytes_per_sample*
                                 genparm[DA_OUTPUT_SPEED])/rx_daout_block;
    while(rx_output_blockrate > ui.max_dma_rate)
      {
      rx_output_blockrate/=2;
      rx_daout_block*=2;
      }
DEB"\nRXout interrupt_rate: Desired %f ",rx_output_blockrate);
    frag=make_dafrag();
    if(ioctl(rx_audio_out,SNDCTL_DSP_SETFRAGMENT,&frag)==-1)
      {
      lirerr(1018);
      return;
      }
#ifdef SNDCTL_DSP_COOKEDMODE
    j=0;
    ioctl(rx_audio_out, SNDCTL_DSP_COOKEDMODE, &j);
#endif
    if(rx_daout_bytes == 1)
      {
      i=AFMT_U8;
      }
    else
      {
      i=AFMT_S16_LE;
      }
    j=i;
    if(ioctl(rx_audio_out, SNDCTL_DSP_SETFMT, &i) == -1)
      {
      lirerr(1020);
      return;
      }
    if(i != j)
      {
      lirerr(1021);
      return;
      }
    i=rx_daout_channels;
    if(ioctl(rx_audio_out, SNDCTL_DSP_CHANNELS, &i) == -1)
      {
      lirerr(1019);
      return;
      }
    if(i != rx_daout_channels)
      {
      lirerr(1099);
      return;
      }
    i=genparm[DA_OUTPUT_SPEED];
    if(i > 0)
      {
      if(ioctl(rx_audio_out, SNDCTL_DSP_SPEED, &i)==-1)
        {
        lirerr(1022);
        return;
        }
      t1=(float)(i)/genparm[DA_OUTPUT_SPEED];
      if(t1 > 1.05 || t1 < 0.95)
        {
        lirerr(1023);
        return;
        }
      genparm[DA_OUTPUT_SPEED]=i;
      }
    rx_da_wrbuf=malloc(rx_daout_block);
    if(rx_da_wrbuf == NULL)
      {
      lirerr(1025);
      return;
      }
    for(i=0; i<rx_daout_block; i++)rx_da_wrbuf[i]=0;
// This has to be the last ioctl call before read or write!!
    ioctl(rx_audio_out, SNDCTL_DSP_GETBLKSIZE,&i);
    lir_rx_dawrite();
    rx_da_info.fragsize=0;
    if(ioctl(rx_audio_out,SNDCTL_DSP_GETOSPACE, &rx_da_info) == -1)lirerr(856432);
    rx_da_totbytes=rx_da_info.fragstotal*rx_da_info.fragsize;
    if(rx_da_info.fragsize != i)
      {
      lirerr(1065);
      return;
      }
    }  
  }
if(rx_da_info.fragsize != rx_daout_block)
  {
  free(rx_da_wrbuf);
  rx_daout_block=rx_da_info.fragsize;
  if(rx_daout_block < daout_size/4)rx_daout_block=daout_size/4;
  rx_da_wrbuf=malloc(rx_daout_block);
  if(rx_da_wrbuf == NULL)
    {
    lirerr(1025);
    return;
    }
  } 
rx_output_blockrate=(float)(rx_daout_bytes_per_sample*
                               genparm[DA_OUTPUT_SPEED])/rx_daout_block;
DEB"  Actual %f  \n",rx_output_blockrate);
}

#define SDI_COL 78

void display_sdi(void)
{
FILE *sdifile;
char *p;
char work_line [80];
int  sdi_screen_line_counter;
char sdi_intro_msg [26]= "SOUND DRIVER INFORMATION:" ;
char sdi_error_msg [46]= "No OSS or ALSA sound-driver information found";
char sndstat_filename [25];
char sndlog_sep [80];
if(screen_last_col < SDI_COL+30)return;
memset (sndlog_sep,'-',sizeof(sndlog_sep));
sndlog_sep[0]='\n';
sndlog_sep[79]='\0';
SNDLOG sndlog_sep);
SNDLOG"\n%s ", sdi_intro_msg );
settextcolor(YELLOW);
lir_text(SDI_COL, 0, sdi_intro_msg );
sdi_screen_line_counter=1;
// Look for OSS-driver info
strcpy (sndstat_filename,"/dev/sndstat");
sdifile=fopen ( sndstat_filename,"r");
// if OSS not active, try  ALSA
if(sdifile==NULL)
  {
  strcpy (sndstat_filename,"/proc/asound/oss/sndstat");
  sdifile=fopen (sndstat_filename,"r");
  }
if(sdifile!=NULL)
  {
  SNDLOG"\n(output of 'cat %s' command): \n\n",sndstat_filename );
  settextcolor(LIGHT_GREEN);
  while(fgets(work_line, sizeof(work_line)-1, sdifile) != NULL && 
                                      sdi_screen_line_counter < screen_last_line)
    {
// check if work_line is ready for output processing
    p=strchr(work_line, '\n');
    if(p != NULL)p[0]=0;
    if(*work_line != '\0')
      {
// display work_line only if it is not blank
// and there are enough lines available on the screen
// and there are at least 40 characters available on a line
      if(sdi_screen_line_counter < screen_last_line &&
                                               (screen_last_col - 40)>= 80)
        {
        sdi_screen_line_counter++;
        lir_text (SDI_COL, sdi_screen_line_counter, work_line);
        }
// write  to soundboard_init.log
      SNDLOG"%s\n",work_line);
      }
    }
  fclose(sdifile);
  }
else
  {
  settextcolor(LIGHT_RED);
  lir_text( SDI_COL, sdi_screen_line_counter++, sdi_error_msg);
  SNDLOG"\n%s",sdi_error_msg );
  }
SNDLOG sndlog_sep);
sndlog_sync();
settextcolor(7);
}

void ossdev_show(char *ss, int no, int minrate, int maxrate, 
                                           int chan, int byte, char *modes)
{
int i, j, k;
sprintf(ss," %02d  %s                 ",no, dev_name);
sprintf(&ss[22],"%d/%d                ",minrate,maxrate);
i=40;
k=1;
while(k < 8)
  {
  if(chan&k)
    {
    sprintf(&ss[i],"%d               ",k);
    i+=2;
    }
  k*=2;
  }  
k=1;
i=49;
while(k < 8)
  {
  if(byte&k)
    {
    sprintf(&ss[i],"%d               ",k);
    i+=2;
    }
  k*=2;
  }  
if(modes == NULL)
  {
  i=57;
  for(j=0; j<3; j++)
    {
    k=1<<j;
    if( (dev_flag[no]&k) != 0)
      {
      sprintf(&ss[i],"%s ",devmode_txt[j]);
      while(ss[i] != 0)i++;
      }
    }
  }
else
  {
  sprintf(&ss[57],"%s ",modes);
  }    
}

void log_output_mode(int no)
{
SNDLOG"\n%s opened as %s ",dev_name, devmode_txt[no]);
sndlog_sync();
}


int select_rdwr_mode(int *line, int no, int mode)
{
int m, k;
char s[80];
if(mode == O_RDONLY)
  {
  k=DEVFLAG_R;
  }
else
  {
  k=DEVFLAG_W;
  }
m=mode;
if( (dev_flag[no]&DEVFLAG_RW) != 0)
  {
  if( (dev_flag[no]&k) != 0)
    {
    line[0]++;
gt_rdo:;
    sprintf(s,"Open %s as RDWR? (Y/N)  =>",dev_name);
    lir_text(10,line[0],s);
    await_processed_keyboard();
    if(kill_all_flag)return 0;
    if(lir_inkey == 'Y')goto rdwr;
    if(lir_inkey != 'N')goto gt_rdo;
    SNDLOG"\nUser selected RDONLY");
    }
  else
    {
rdwr:;
    m=O_RDWR;
    SNDLOG"\nUser selected RDWR");
    }
  line[0]++;
  }
else
  {
  SNDLOG"\nRDWR not available");
  }
return m;
}

void select_ossdev(int *line, int sound_type)
{
int iohandle;
int caps;
int i, j, k, m, n, nn;
char ss[3*MAX_COLUMN];
int mode, rdwr_fmt, device_no;
int rdwr_min_speed;
int rdwr_max_speed;
audio_buf_info ad_info;
char *rdtest_buffer;
int rdwr_channels, suggested_dev;
int dev_rdwr_channels[MAX_DEVNAMES];
int dev_max_rdwr_speed[MAX_DEVNAMES];
int dev_min_rdwr_speed[MAX_DEVNAMES];
int dev_rdwr_bytes[MAX_DEVNAMES];
int column, maxcolumn, bottom_line;
int devflag_val;
char s[256];
char color;
int h_test;
int h_test2;
h_test=-1;
h_test2=-1;
rdtest_buffer=malloc(2048);
if(rdtest_buffer == NULL)
  {
  lirerr(1044);
  return;
  }
n=0;
suggested_dev=-1;
maxcolumn=screen_last_xpixel/text_width-1;
if(maxcolumn > MAX_COLUMN)maxcolumn=MAX_COLUMN;
bottom_line=0;
line[0]++;
sprintf(s,"Dev     device          sample-rate    channels  bytes       modes");
SNDLOG"%s\n",s);
lir_text(0,line[0],s);
line[0]++;
sprintf(s," no      file           min / max");
SNDLOG"%s\n",s);
lir_text(0,line[0],s);
line[0]++;
sprintf(s," --  ---------------  --------------   -------  -------  -------------------" );
SNDLOG"%s\n",s);
lir_text(0,line[0],s);
line[0]++;
for(device_no=0; device_no<MAX_DEVNAMES; device_no++)
  {
  lir_sched_yield();
  dev_flag[device_no]=0;
  make_devname(device_no, dev_name);
  SNDLOG"Checking %s\n",dev_name);
  if(device_no == (ui.rx_addev_no&255) && ui.rx_admode == O_RDWR)
    {
    if(ui.rx_dadev_no != device_no && ui.tx_dadev_no != device_no)
      {
      k=2;
      if( (ui.rx_input_mode&DWORD_INPUT) )k=4;
      ossdev_show(ss, device_no, ui.rx_ad_speed, ui.rx_ad_speed,
                      ui.rx_ad_channels, k,"RDWR used by RXin");
      settextcolor(10);
      line[0]++;
      lir_text(0,line[0],ss);
      lir_refresh_screen();
      dev_flag[device_no]=DEVFLAG_RW;
      }
    else
      {
      dev_flag[device_no]=0;
      }    
    goto used_rdwr;      
    
    }
  if(device_no == ui.rx_dadev_no && ui.rx_damode == O_RDWR)
    {
    k=2;
    if( (ui.rx_input_mode&DWORD_INPUT) )k=4;
    ossdev_show(ss, device_no, ui.rx_ad_speed, ui.rx_ad_speed,
                      ui.rx_ad_channels, k,"RDWR used by RXout");
    settextcolor(10);
    line[0]++;
    lir_text(0,line[0],ss);
    lir_refresh_screen();
    dev_flag[device_no]=DEVFLAG_RW;
    goto used_rdwr;      
    }
  sndlog_sync();
  for(mode=0; mode<3; mode++)
    {
    devflag_val=1<<mode;
    sprintf(s,
    "Trying to open %s %s. Device defective if system hangs here. Read %s   ",
    dev_name, devmode_txt[mode],rx_logfile_name);
    sndlog_sync();
    settextcolor(12);
    lir_text(0,line[0]+1,s);
    lir_refresh_screen();
    settextcolor(7);
    iohandle=open( dev_name, devmodes[mode]|O_NONBLOCK, 0);
    clear_lines(line[0]+1,line[0]+1);
    if(iohandle != -1)
      {
#ifdef SNDCTL_DSP_COOKEDMODE
      i=0;
      ioctl(iohandle, SNDCTL_DSP_COOKEDMODE, &i);
#endif
      for(nn=0; nn<80; nn++)SNDLOG"-");
      log_output_mode(mode);
      rdwr_fmt=0;
      rdwr_channels=0;
// Check for 16 bits
      if(ioctl(iohandle, SNDCTL_DSP_GETFMTS, &m) == -1)
        {
        SNDLOG"\nioctl(GETFMTS)  failed");
baddev:;
        sndlog_sync();
        devflag_val=0;
        goto cls;
        }
      if( (m&AFMT_U8) ==0)
        { 
        SNDLOG"\n8bit format NOT supported");
        rdwr_fmt|=1;
        }
      if( (m&AFMT_S16_LE) == 0)
        {
        SNDLOG"\n16bit format NOT supported");
        }
      else
        {
        SNDLOG"\n16bit format supported");
        rdwr_fmt|=2;
        }
#ifdef AFMT_S32_LE
// Check for 32 bits
      if( (m&AFMT_S32_LE) != 0)
        {
        rdwr_fmt|=4;
        devflag_val=1<<mode;
        SNDLOG"\n32bit format supported");
        }
      else
        {
        SNDLOG"\n32bit format NOT supported");
        }
#endif
      sndlog_sync();
      if(rdwr_fmt == 0)
        {
        devflag_val=0;
        goto cls;
        }
// Find max no of channels.
// We never want more than 4 input channels!!
      rdwr_channels=0;
      m=4;
      while(m > 0)
        {
        k=m;
        j=ioctl(iohandle, SNDCTL_DSP_CHANNELS, &m);
        if(j != -1 && k == m)rdwr_channels|=m;  
        m/=2;
        }
      if(rdwr_channels == 0)
        {
        k=0;
        SNDLOG"\nMax no of channels = 0");
        goto baddev;
        }
      else
        {
        SNDLOG"\nChannels = %d",rdwr_channels);
        }
      sndlog_sync();
//
// To allow a separate #ifdef __FreeBSD__
// - Diane Bruce VA3DB, Sept 6, 2006
//
#ifdef __linux__
// Find maximum speed
      rdwr_max_speed=ABOVE_MAX_SPEED;
      if(ioctl(iohandle, SNDCTL_DSP_SPEED, &rdwr_max_speed)==-1)
        {
        SNDLOG"\nioctl(SPEED)  failed (max)");
        goto baddev;
        }
      SNDLOG"\nMax speed %d Hz",rdwr_max_speed);
// Find minimum speed
      i=0;
      while(i<MAX_LOWSPEED)
        {
        rdwr_min_speed=low_speeds[i];
        if(ioctl(iohandle, SNDCTL_DSP_SPEED, &rdwr_min_speed)!=-1)
                                                                  goto minok;
        i++;
        }
      lirerr(1040);
      goto sel_ossdev_skip;
minok:;
#endif
//
// This code finds min max speed of sound card for FreeBSD
// It probably works for other BSDs as well
// Diane Bruce, db@db.net VA3DB
// Sept 7, 2006
//
#ifdef __FreeBSD__
        {
        snd_capabilities capab;
        int status;
        bzero((void *)&capab, sizeof(capab));
        status=ioctl(iohandle, AIOGCAP, &capab);
        if(status < 0)goto sel_ossdev_skip;
        rdwr_min_speed=capab.rate_min;
        rdwr_max_speed=capab.rate_max;
        SNDLOG"\nMax output speed %d", rdwr_max_speed);
        }
#endif
      SNDLOG"\nMin sampling speed = %d Hz",rdwr_min_speed);
      sndlog_sync();
      if(sound_type == RXAD)
        {
//devmode_txt[3]={"RDONLY","WRONLY","RDWR"};
        if(mode != DEVMODE_WRONLY)
          {
          if(ioctl(iohandle, SNDCTL_DSP_GETISPACE, &ad_info) == -1)
            {
            SNDLOG"\nioctl(GETISPACE)  failed");
            goto baddev;
            }
// If it is opened RDWR, check we have a chance to write while reading
// Some devices allow write or read in rdwr mode - but not simultaneously.
          if(mode == DEVMODE_RDWR)
            {
            read(iohandle, rdtest_buffer, 128);
            if(ioctl(iohandle,SNDCTL_DSP_GETOSPACE, &ad_info) == -1)
              {
              SNDLOG"\nioctl(GETOSPACE)  failed while reading");
              goto baddev;
              }
            }
          }  
        }
cls:;
      if(close(iohandle)==-1)
        {
        sndlog_sync();
        lirerr(1037);
        goto sel_ossdev_skip;
        }
      if(devflag_val != 0)
        {
        if(sound_type == RXAD || sound_type == TXAD)
          {      
          if(devflag_val != 2)
            {
            suggested_dev=device_no;
            n++;
            }
          }  
        if(sound_type == RXDA || sound_type == TXDA)
          {      
          if(devflag_val != 1)
            {
            suggested_dev=device_no;
            n++;
            }
          }  
        dev_flag[device_no]|=devflag_val;
        switch (mode)
          {
          case DEVMODE_RDONLY:
          dev_rd_channels[device_no]=rdwr_channels;
          dev_max_rd_speed[device_no]=rdwr_max_speed;
          dev_min_rd_speed[device_no]=rdwr_min_speed;
          dev_rd_bytes[device_no]=rdwr_fmt;
          break;
    
          case DEVMODE_WRONLY:
          dev_wr_channels[device_no]=rdwr_channels;
          dev_max_wr_speed[device_no]=rdwr_max_speed;
          dev_min_wr_speed[device_no]=rdwr_min_speed;
          dev_wr_bytes[device_no]=rdwr_fmt;
          break;
    
          case DEVMODE_RDWR:
          dev_rdwr_channels[device_no]=rdwr_channels;
          dev_max_rdwr_speed[device_no]=rdwr_max_speed;
          dev_min_rdwr_speed[device_no]=rdwr_min_speed;
          dev_rdwr_bytes[device_no]=rdwr_fmt;
          break;
          }
        }
      SNDLOG"\n");
      }
    }
  if(dev_flag[device_no] != 0)
    {
    SNDLOG"\nFLAG= %d",dev_flag[device_no]);
// One device may be different hardwares for input and output.
// Put them on different lines if characteristics differ.
    k=0;
    j=0;
// Store equal pairs in k
// bit 0=1 => READ==WRITE
// bit 1=1 => READ==RDWR
// bit 2=1 => WRITE==RDWR
    if( (dev_flag[device_no]&(DEVFLAG_R+DEVFLAG_W)) == DEVFLAG_R+DEVFLAG_W)
      {
      j|=1;
      if(dev_rd_channels[device_no]==dev_wr_channels[device_no] &&
         dev_max_rd_speed[device_no]==dev_max_wr_speed[device_no] &&
         dev_min_rd_speed[device_no]==dev_min_wr_speed[device_no] &&
         dev_rd_bytes[device_no]==dev_wr_bytes[device_no]) k|=1;
      }
    if( (dev_flag[device_no]&(DEVFLAG_R+DEVFLAG_RW)) == DEVFLAG_R+DEVFLAG_RW)
      {
      j|=2;
      if(dev_rd_channels[device_no]==dev_rdwr_channels[device_no] &&
         dev_max_rd_speed[device_no]==dev_max_rdwr_speed[device_no] &&
         dev_min_rd_speed[device_no]==dev_min_rdwr_speed[device_no] &&
         dev_rd_bytes[device_no]==dev_rdwr_bytes[device_no])k|=2;
      }
    if( (dev_flag[device_no]&(DEVFLAG_W+DEVFLAG_RW)) == DEVFLAG_W+DEVFLAG_RW)
      {
      j|=4;
      if(dev_wr_channels[device_no]==dev_rdwr_channels[device_no] &&
         dev_max_wr_speed[device_no]==dev_max_rdwr_speed[device_no] &&
         dev_min_wr_speed[device_no]==dev_min_rdwr_speed[device_no] &&
         dev_wr_bytes[device_no]==dev_rdwr_bytes[device_no])k|=4;
      }
    SNDLOG"\nj=%d   k=%d",j,k);
    if(j == k)
      {
// There are no differences between devices. Put everything on one line.

      m=1;
      if( (dev_flag[device_no]&DEVFLAG_R) == DEVFLAG_R)
        {
        ossdev_show(ss, device_no,
                        dev_min_rd_speed[device_no],
                        dev_max_rd_speed[device_no],
                        dev_rd_channels[device_no],
                        dev_rd_bytes[device_no],NULL);
        }
      else
        {
        if( (dev_flag[device_no]&DEVFLAG_W) == DEVFLAG_W)
          {
          ossdev_show(ss, device_no,
                         dev_min_wr_speed[device_no],
                         dev_max_wr_speed[device_no],
                         dev_wr_channels[device_no],
                         dev_wr_bytes[device_no],NULL);
          }
        }
      if(sound_type == RXAD || sound_type == TXAD)
        {      
        if( (dev_flag[device_no]&(DEVFLAG_R+DEVFLAG_RW)) == 0)
          {
          color=0x17;
          }
        else
          {
          color=0x1c;
          }
        }
      else
        {
        if( (dev_flag[device_no]&(DEVFLAG_W+DEVFLAG_RW)) == 0)
          {
          color=0x17;
          }
        else
          {
          color=0x1c;
          }
        }    
      }
    else
      {
// There are differences.
// We can not arrive here unless there are at least two modes
// and consequently j can not be zero.
      if( (j&1) == 0)
        {
// We do not have both read and write so it is one of them plus rdwr.
// This does not seem right, but this code although untested will hopefully
// give meaningful information in case we ever arrive here.
        lirerr(86765);
        goto sel_ossdev_skip;
        }
      else
        {
// There is both read and write.
        if( (dev_flag[device_no]&DEVFLAG_RW) == 0)
          {
// but there is no rdwr
          m=2;
          color=0x17;
          ossdev_show(&ss[maxcolumn], device_no,
                          dev_min_wr_speed[device_no],
                          dev_max_wr_speed[device_no],
                          dev_wr_channels[device_no],
                          dev_wr_bytes[device_no],devmode_txt[1]);
          ossdev_show(ss, device_no,
                          dev_min_rd_speed[device_no],
                          dev_max_rd_speed[device_no],
                          dev_rd_channels[device_no],
                          dev_rd_bytes[device_no],devmode_txt[0]);
          }
        else
          {
// There is rdwr, read and write.
          if(k == 0)
            {
// They are all different
            m=3;
            color=0x17;
            ossdev_show(&ss[2*maxcolumn], device_no,
                          dev_min_wr_speed[device_no],
                          dev_max_wr_speed[device_no],
                          dev_wr_channels[device_no],
                          dev_wr_bytes[device_no],devmode_txt[1]);
            ossdev_show(&ss[maxcolumn], device_no,
                          dev_min_rd_speed[device_no],
                          dev_max_rd_speed[device_no],
                          dev_rd_channels[device_no],
                          dev_rd_bytes[device_no],devmode_txt[0]);
            ossdev_show(ss, device_no,
                          dev_min_rdwr_speed[device_no],
                          dev_max_rdwr_speed[device_no],
                          dev_rdwr_channels[device_no],
                          dev_rdwr_bytes[device_no],devmode_txt[2]);
            }
          else
            {
// rdwr equals read or write.
            m=2;
            if( (k&4) == 0)
              {
              color=0x17;
              ossdev_show(&ss[maxcolumn], device_no,
                           dev_min_wr_speed[device_no],
                           dev_max_wr_speed[device_no],
                           dev_wr_channels[device_no],
                           dev_wr_bytes[device_no],devmode_txt[1]);
              ossdev_show(ss, device_no,
                           dev_min_rd_speed[device_no],
                           dev_max_rd_speed[device_no],
                           dev_rd_channels[device_no],
                           dev_rd_bytes[device_no],
                           devmode_txt[0]);
              strcat(ss,devmode_txt[2]);
              }
            else
              {
              color=0x17;
              ossdev_show(&ss[maxcolumn], device_no,
                      dev_min_wr_speed[device_no],
                      dev_max_wr_speed[device_no],
                      dev_wr_channels[device_no], dev_wr_bytes[device_no],
                      devmode_txt[1]);
              strcat(ss,devmode_txt[2]);
              ossdev_show(ss, device_no,
                      dev_min_rd_speed[device_no],
                      dev_max_rd_speed[device_no],
                      dev_rd_channels[device_no],
                      dev_rd_bytes[device_no],
                      devmode_txt[0]);
              }
            }
          }
        }
      }
    while (m >0)
      {
      m--;
      line[0]++;
      if(line[0] >= screen_last_line-1)
        {
        for(column=0; column<maxcolumn; column++)s[column]=' ';
        s[maxcolumn]=0;
        lir_text(0,line[0],s);
        settextcolor(14);
        lir_text(10,line[0],"(press ENTER to continue)");
        settextcolor(7);
        await_keyboard();
        if(kill_all_flag) goto sel_ossdev_skip;
        lir_text(0,line[0],s);
        bottom_line=screen_last_line-1;
        line[0]=1;
        }
      settextcolor(color);
      column=0;
      while(ss[m*maxcolumn+column]!=0)column++;
      while(column < maxcolumn)
        {
        ss[m*maxcolumn+column]=' ';
        column++;
        }
      ss[(m+1)*maxcolumn-1]=0;
      lir_text(0,line[0],&ss[m*maxcolumn]);
      lir_refresh_screen();
      SNDLOG"\n%s    color=%d m=%d",&ss[m*maxcolumn],color,m);
      color=0x1c;
      }
    SNDLOG"\n");
    }
used_rdwr:;  
  }
settextcolor(7);
line[0]++;
clear_lines(line[0],line[0]);
if(bottom_line > line[0])line[0]=bottom_line;
display_sdi();
if(sound_type == TXDA)
  {      
  if(n == 0)
    {
    ui.tx_dadev_no=-1;
    lir_text(0,line[0],"No device found for Tx output");
    lir_text(2,line[0]+1,"ERROR (press any key)");
    await_keyboard();
    goto sel_ossdev_skip;
    }
  settextcolor(14);
  lir_text(0,line[0],"Select device for Tx output");
get_tx_dadev:;
  settextcolor(7);
  ui.tx_dadev_no=lir_get_integer(29, line[0], 3, 0,MAX_DEVNAMES-1);
  if(kill_all_flag) goto sel_ossdev_skip;
  if( (dev_flag[ui.tx_dadev_no]&(DEVFLAG_W+DEVFLAG_RW)) == 0)
    {
    settextcolor(15);
    lir_text(29,line[0],"???");
    lir_text(25,line[0]+1,"ERROR (press any key)");
    await_keyboard();
    if(kill_all_flag) goto sel_ossdev_skip;
    clear_lines(line[0]+1,line[0]+1);
    goto get_tx_dadev;
    }
  line[0]+=2;
  }
if(sound_type == RXDA)
  {      
  if(n == 0)
    {
    if(soundcard_test_cmd_flag[RXAD] == THRFLAG_ACTIVE)goto sel_ossdev_skip;
    if(line[0] > screen_last_line-6)line[0]=screen_last_line-6;
    lir_text(0,line[0],
           "No device found for 1/2 byte output while input is open.");
    line++;
    lir_text(0,line[0],
           "You may use Linrad to watch spectra and record raw data.");
    line++;
    lir_text(0,line[0],
           "Linrad will now look for devices that can be used to");
    line++;
    lir_text(0,line[0],
           "output sound while processing recorded data.");
    line++;
    lir_text(0,line[0],press_any_key);
    SNDLOG"\nNo output device found while input is open.");
    await_keyboard();
    goto sel_ossdev_skip;
    }
  if(n==1)ui.rx_dadev_no=suggested_dev;  
  if(n > 1)
    {
get_rx_dadev:;
    settextcolor(14);
    lir_text(0,line[0],"Select device for RX output by Dev. no.");
    display_sdi();
    settextcolor(7);
    ui.rx_dadev_no=lir_get_integer(40, line[0], 3, 0,MAX_DEVNAMES-1);
    if(kill_all_flag)goto sel_ossdev_skip;
    if( (dev_flag[ui.rx_addev_no]&(DEVFLAG_W+DEVFLAG_RW)) == 0)
      {
      settextcolor(15);
      lir_text(51,line[0],"???");
      lir_text(35,line[0]+1,"ERROR (press any key)");
      await_keyboard();
      if(kill_all_flag) goto sel_ossdev_skip;
      clear_lines(line[0],line[0]+1);
      goto get_rx_dadev;
      }
    }
  goto sel_ossdev_skip;
  }
if(sound_type == TXAD)
  {      
  if(n == 0)
    {
    lir_text(0,line[0],"No device found for microphone input.");
    line[0]++;
    lir_text(2,line[0]+1,"ERROR (press any key)");
    await_keyboard();
    goto sel_ossdev_skip;
    }
  settextcolor(14);
  lir_text(0,line[0],"Select device for microphone input");
  get_micdev:;
  settextcolor(7);
  ui.tx_addev_no=lir_get_integer(37, line[0], 3, 0,MAX_DEVNAMES-1);
  if( (dev_flag[ui.tx_addev_no]&(DEVFLAG_R+DEVFLAG_RW)) == 0)
    {
    settextcolor(15);
    lir_text(37,line[0],"???");
    lir_text(25,line[0]+1,"ERROR (press any key)");
    await_keyboard();
    if(kill_all_flag)goto sel_ossdev_skip;
    clear_lines(line[0]+1,line[0]+1);
    goto get_micdev;
    }
  goto sel_ossdev_skip;
  }
if(sound_type == RXAD)
  {      
  if(n == 0)
    {
    if(line[0] > screen_last_line-13)line[0]=screen_last_line-13;
    lir_text(0,line[0],"No sound device found (for 2/4 byte input).");
    line[0]+=2;
    lir_text(10,line[0],"If you are using OSS:");
    line[0]++;
    lir_text(0,line[0],"Run soundconf command to make sure OSS is correctly");
    line[0]++;
    lir_text(0,line[0],"initialised for your sound board(s).");
    line[0]++;
    lir_text(0,line[0],"Then start sound with soundon command.");
    line[0]+=2;
    lir_text(10,line[0],"Without OSS:");
    line[0]++;
    lir_text(0,line[0],
                 "Find out how to configure your sound device drivers(s)");
    line[0]++;
    lir_text(0,line[0],"Some distributions have the sndconfig command.");
    line[0]++;
    lir_text(0,line[0],
          "Soundcard drivers are usually configured during Linux install");
    line[0]+=2;
    lir_text(20,line[0],press_any_key);
    await_keyboard();
    SNDLOG"\nNo input device found");
    }
  if(n==1)ui.rx_addev_no=suggested_dev;  
  if(n > 1)
    {
get_rx_addev:;
    settextcolor(14);
    lir_text(0,line[0],"Select device for RX input by Dev. no.");
    display_sdi();
    settextcolor(7);
    ui.rx_addev_no=lir_get_integer(39, line[0], 3, 0,MAX_DEVNAMES-1);
    if(kill_all_flag)goto sel_ossdev_skip;
    if( (dev_flag[ui.rx_addev_no]&(DEVFLAG_R+DEVFLAG_RW)) == 0)
      {
      settextcolor(15);
      lir_text(51,line[0],"???");
      lir_text(35,line[0]+1,"ERROR (press any key)");
      await_keyboard();
      if(kill_all_flag) goto sel_ossdev_skip;
      clear_lines(line[0],line[0]+1);
      goto get_rx_addev;
      }
    }
  make_devname(ui.rx_addev_no, dev_name);
  SNDLOG"\n\n %s selected by user for input.",dev_name);
  SNDLOG"   dev_flag=%d",dev_flag[ui.rx_addev_no]);
  sndlog_sync();
// High end users might need two devices to read four A/D channels.
// This is an option introduced April 2004 to allow two receive
// channels with the Lynx Two card (model A).
// Do not confuse other users by asking for a second input device
// unless the selected device has 32 bits but only allows 2 channels.
  if(dev_rd_bytes[ui.rx_addev_no] >= 4 &&
     dev_rd_channels[ui.rx_addev_no]==2 &&
     (dev_flag[ui.rx_addev_no]&DEVFLAG_R) != 0)
    {
    SNDLOG"\n\nUser selected 4 byte device with max two channels.");
    j=0;
    line[0]=1;
    for(i=0; i<MAX_DEVNAMES; i++)
      {
      if(i != ui.rx_addev_no &&
         dev_max_rd_speed[i] == dev_max_rd_speed[ui.rx_addev_no] &&
         dev_rd_bytes[i] >= 4 &&
         (dev_flag[i]&DEVFLAG_R) != 0)
        {
// This device can also read 32 bits at the same speed.
// Check if it can be opened simultaneously in read mode.
        if(j==0)
          {
          line[0]++;
          lir_text(5,line[0],"This device allows two channels only.");
          line[0]++;
          lir_text(5,line[0],"Do you want to use a second device to run two");
          line[0]++;
          lir_text(5,line[0],"receive channels with four audio channels ?");
          line[0]++;
          lir_text(5,line[0],"(Y/N)?");
          line[0]++;
gt_dual:;
          await_processed_keyboard();
          if(kill_all_flag) goto sel_ossdev_skip;
          if(lir_inkey == 'N')
            {
            SNDLOG"\nUser skipped dual devices");
            goto skip_dual_devices;
            }
          if(lir_inkey != 'Y')goto gt_dual;
          SNDLOG"\nUser wants to try to open two read devices");
          make_devname(ui.rx_addev_no, dev_name);
          h_test=open( dev_name, O_RDONLY, 0);
          if(h_test == -1)
            {
            lirerr(1102);
            goto sel_ossdev_skip;
            }
#ifdef SNDCTL_DSP_COOKEDMODE
          k=0;
          ioctl(h_test, SNDCTL_DSP_COOKEDMODE, &k);
#endif
// Check for DSP_CAP_TRIGGER capability
          if(ioctl(h_test, SNDCTL_DSP_GETCAPS,&caps) == -1)
            {
            lirerr(1182);
            goto sel_ossdev_skip;
            }
          if( (caps&DSP_CAP_TRIGGER) == 0)
            {
            SNDLOG"\nDevice does not have DSP_CAP_TRIGGER");
            close(h_test);
            goto skip_dual_devices;
            }
          j=2;
          if(ioctl(h_test, SNDCTL_DSP_CHANNELS, &j) == -1)
            {
            lirerr(1185);
            goto sel_ossdev_skip;
            }
          if(j != 2)
            {
            lirerr(1186);
            goto sel_ossdev_skip;
            }
          }
        make_devname(i, dev_name);
        h_test2=open( dev_name, O_RDONLY|O_NONBLOCK, 0);
        if(h_test2 != -1)
          {
#ifdef SNDCTL_DSP_COOKEDMODE
          k=0;
          ioctl(h_test2, SNDCTL_DSP_COOKEDMODE, &k);
#endif
          line[0]++;
          SNDLOG"\n%s opened as second read device",dev_name);
          sprintf(s,"%2d:%12s  %7d - %7d Hz   %d Chan.   %d bit",
                    i,dev_name,
                    dev_min_rd_speed[i],
                    dev_max_rd_speed[i],
                    dev_rd_channels[i],
                    dev_rd_bytes[i]);
          lir_text(0,line[0],s);
          close(h_test2);
          dev_flag[i]|=256;
          }
        }
      }
    settextcolor(14);
    lir_text(0,line[0]+1,"Select second input device by Dev no.");
get_addev2:;
    settextcolor(7);
    i=lir_get_integer(51, line[0]+1, 3, 0,MAX_DEVNAMES-1);
    if(kill_all_flag)goto sel_ossdev_skip;
    if( (256 & dev_flag[i]) == 0)
      {
      settextcolor(15);
      lir_text(40,line[0],"???");
      lir_text(35,line[0]+1,"ERROR (press any key)");
      await_keyboard();
      if(kill_all_flag) goto sel_ossdev_skip;
      clear_lines(line[0]+1,line[0]+1);
      settextcolor(7);
      goto get_addev2;
      }
// We store the second A/D device in the second byte of ad_device_no
    make_devname(i, dev_name);
    ui.rx_addev_no+=256*(i+1);
    dev_flag[i]&=255;
    h_test2=open( dev_name, O_RDONLY|O_NONBLOCK, 0);
    if(h_test2 == -1)
      {
      lirerr(1172);
      goto sel_ossdev_skip;
      }
#ifdef SNDCTL_DSP_COOKEDMODE
    k=0;
    ioctl(h_test2, SNDCTL_DSP_COOKEDMODE, &k);
#endif
// Check for DSP_CAP_TRIGGER capability
    if(ioctl(h_test2, SNDCTL_DSP_GETCAPS,&caps) == -1)
      {
      lirerr(1182);
      goto sel_ossdev_skip;
      }
    if( (caps&DSP_CAP_TRIGGER) == 0)
      {
      SNDLOG"\nSecond input device does not have DSP_CAP_TRIGGER");
      close(h_test);
      close(h_test2);
      goto skip_dual_devices;
      }
    SNDLOG"\n %s is second input device.",dev_name);
    ui.rx_input_mode=TWO_CHANNELS+IQ_DATA;
    ui.rx_rf_channels=2;
    ui.rx_ad_channels=4;
    clear_screen();
    goto continue_dual;
skip_dual_devices:;
    clear_screen();
    }
// Now that we decided what device to use for input, open it and let
// the user decide how to use it.
  if(ui.rx_addev_no != -1)
    {
    make_devname(ui.rx_addev_no, dev_name);
    sndlog_sync();
    line[0]++;
    ui.rx_admode=select_rdwr_mode(line, ui.rx_addev_no, O_RDONLY);
    if(kill_all_flag) goto sel_ossdev_skip;
    sndlog_sync();
    }
  clear_screen();
  line[0]=0;
  mode=0;
  if(ui.rx_admode == O_RDWR)mode=2;
  sprintf(s,"For analog input: %s opened in %s mode",
                                               dev_name,devmode_txt[mode]);
  lir_text(0,line[0],s);
  line[0]+=2;
  rx_audio_in=open( dev_name, ui.rx_admode, 0);
  if(rx_audio_in == -1)
    {
    lirerr(1038);
    goto sel_ossdev_skip;
    }
continue_dual:;
  clear_lines(10,11+j);
#ifdef SOUND_OSS
  sprintf(s,"Use ossmix");
#else
  sprintf(s,"Use system mixer program");
#endif
  column=0;
  while(s[column] != 0)column++;
  sprintf(&s[column]," to select input and");
  lir_text(0,line[0],s);
  line[0]++;
  lir_text(0,line[0],"to disable direct connection from input to output etc.");
  line[0]+=2;
  lir_text(0,line[0],"In case selection of sampling frequencies is not what you");
  line[0]++;
  lir_text(0,line[0],
        "expect, check README files for your board - there may be ways");
  line[0]++;
  lir_text(0,line[0],"to reconfigure the hardware. (Under OSS, run ossmix)");
  }
sel_ossdev_skip:;
free(rdtest_buffer);
line[0]+=2;
}

void set_rx_dadev_parms(int *line)
{
ui.rx_min_da_channels=1;
if( (dev_wr_channels[ui.rx_dadev_no]&3) == 0)
  {
  lir_text(line[0],0,
  "Device must have 1 or 2 channels (or both).");
  line[0]++;
  ui.rx_dadev_no=-1;
  lir_text(line[0],10,press_any_key);
  await_keyboard();
  return;
  }      
ui.rx_min_da_channels=1;  
if( (dev_wr_channels[ui.rx_dadev_no]&1) == 0)ui.rx_min_da_channels=2;
ui.rx_max_da_channels=2;
if( (dev_wr_channels[ui.rx_dadev_no]&2) == 0)ui.rx_max_da_channels=1;
ui.rx_min_da_speed=dev_min_wr_speed[ui.rx_dadev_no];
ui.rx_max_da_speed=dev_max_wr_speed[ui.rx_dadev_no];
ui.rx_min_da_bytes=1;  
if( (dev_wr_bytes[ui.rx_dadev_no]&1) == 0)ui.rx_min_da_bytes=2;
ui.rx_max_da_bytes=ui.rx_min_da_bytes;
if( (dev_wr_bytes[ui.rx_dadev_no]&2) == 0)ui.rx_max_da_bytes=1;
}

void stop_iotest(int mode)
{
soundcard_test_cmd_flag[mode]=THRFLAG_KILL;
pthread_join(thread_identifier_soundcard_test[mode],0);
}

int start_iotest(int *line, int mode)
{
char *mode_text[3]={"RX INPUT", "RX OUTPUT", "TX OUTPUT"};
char s[80];
int i;
// To make sure that the devices we select for tx or for rx output can 
// be used while the rx input is running we start a separate thread that
// uses the rx input so we will fail on devices that do not
// allow simultaneous read and write.
soundcard_test_block_count[mode]=0;
soundcard_test_cmd_flag[mode]=THRFLAG_ACTIVE;
pthread_create(&thread_identifier_soundcard_test[mode],
                                   NULL,(void*)thread_iotest[mode], NULL);
i=0;
while(i < 100 && soundcard_test_block_count[mode] < 8)
  {
  lir_sleep(10000);
  if(kill_all_flag)return FALSE;
  i++;
  }
if(i >= 100)
  {
  line[0]+=2;
  settextcolor(12);
  sprintf(s,"FAILED TO RUN THE %s",mode_text[mode]);
  lir_text(0,line[0],s);
  line[0]++;
  if(mode == RXAD)
    {
    lir_text(0,line[0],"Set a working input or force no soundcard by");  
    line[0]++;
    lir_text(0,line[0],"selecting network input, then soundcard select.");  
    line[0]++;
    lir_text(0,line[0],press_any_key);  
    await_keyboard();
    return FALSE;
    }
  }
return TRUE;
}



void set_rx_io(void)
{
char *tmptxt;
int toggle;
int speed_warning;
int dev_max_channels;
int dev_min_channels;
int i, j, k, min, max;
int trySpeed, deltaSpeed;
int rdwr_max_speed;
int err, line, refline, column;
char s[256];
char color;
memset(blanks, ' ',58);
blanks[59]=0;
toggle=TRUE;
color=7;
column=0;
speed_warning=0;
sndlog=fopen(rx_logfile_name, "w");
if(sndlog == NULL)
  {
  lirerr(1016);
  return;
  }
write_log=TRUE;  
if(ui.rx_ad_channels == 0)
  {
  ui.rx_addev_no=-1;
  ui.use_alsa=FALSE;
  }  
if(ui.rx_max_da_channels == 0)
  {
  ui.rx_dadev_no=-1;
  ui.use_alsa=FALSE;
  }  
begin_rx_set_io:;
// Start by writing what sound system we use and allow the user
// to toggle between alsa-oss and native ALSA in case there is a choice.
// Set the previous selection non-valid in case the system is changed.
clear_screen();
#ifdef ALSA_PRESENT
lir_text(0,0,"The current setting is:");
lir_text(0,1,"This system can be set up for native ALSA or for alsa-oss.");
i=ui.use_alsa;
set_syst:;
settextcolor(10);
if(ui.use_alsa)
  {
  lir_text(24,0,"native ALSA");
  }
else
  {
  lir_text(24,0,"alsa-oss   ");
  }
sprintf(s,"ALSA library version:%s",SND_LIB_VERSION_STR);
settextcolor(7);  
lir_text(37,0,s);
if(toggle)
  {
  lir_text(5,3,"Press 'T' to toggle sound interface.");
  lir_text(5,4,"Any other key to continue.");
  toupper_await_keyboard();
  if(kill_all_flag)goto set_rx_io_errexit;
  if(lir_inkey == 'T')
    {
    ui.use_alsa ^=1;
    goto set_syst;
    }
  if(i != ui.use_alsa)
    {
    ui.rx_dadev_no=-1;
    ui.rx_addev_no=-1;
    ui.tx_dadev_no=-1;
    ui.tx_addev_no=-1;
    }
  toggle=FALSE;
  }
#else
ui.use_alsa=FALSE;
lir_text(0,0,"This system uses OSS, alsa-oss or legacy sound.");
#endif
clear_lines(1,4);
line=2;
if((ui.network_flag&NET_RX_INPUT) != 0)
  {
  lir_text(0,line,"Input from the network is currently selected.");
  line++;
  }
// **************************************************************
// Show the current RX input settings on the screen.
// **************************************************************
lir_text(0,line,"Linrad RX input from:");
settextcolor(10);
sprintf(s,"UNKNOWN (no=%d)",ui.rx_addev_no);
tmptxt=s;
i=0;
j=1;
if(ui.rx_addev_no == -1)
  {
  tmptxt="NOT YET SELECTED";
  j=0;
  }
else
  {
  if(ui.rx_addev_no >= SPECIFIC_DEVICE_CODES)
    {
    switch (ui.rx_addev_no)
      {
      case SDR14_DEVICE_CODE:
      tmptxt="SDR-14";
      break;

      case SDRIQ_DEVICE_CODE:
      tmptxt="SDR-IQ";
      break;

      case PERSEUS_DEVICE_CODE:
      tmptxt="PERSEUS";
      break;

      default:
      j=0;
      break;
      }
    }
  else
    {
#ifdef ALSA_PRESENT
    if(ui.use_alsa)
      {
      err=alsa_get_dev_id (ui.rx_addev_no);
      if(err < 0)
        {
	sprintf(s,"invalid card/device  ( alsa-device-seq-nmbr = %i )",
                                                            ui.rx_addev_no );
        tmptxt=s;
        j=0;
	}
      else
        {
	sprintf(s,"Soundcard =%s",alsa_dev_soundcard_name);
        lir_text(22,line,s);
        settextcolor(7);
        i++;
        SNDLOG"Initial RXin setting: %s\n",s);
        sprintf(s,"Dev name = %s",alsa_dev_name);
        lir_text(2,line+i,s);
        i++;
        SNDLOG"%s\n",s);
        sprintf(s,"Pcm name = %s",alsa_dev_plughw_pcm_name);
        lir_text(2,line+i,s);
        i++;
        SNDLOG"%s\n",s);
        sprintf(s,"Seq nmbr = %i",alsa_dev_seq_nmbr);
        lir_text(2,line+i,s);
        SNDLOG"%s\n",s);
	}
      }
    else
#endif
      {
      make_devname(ui.rx_addev_no, dev_name);
      sprintf(s,"Device name =%s",dev_name);
      tmptxt=s;
      }
    }
  }        
if(i == 0)
  {
  lir_text(22,line,tmptxt);
  settextcolor(7);
  SNDLOG"Initial RXin setting: %s\n",tmptxt);
  }
line+=i+1;
if(j)
  {
  sprintf(s,"sample rate       = %i ",ui.rx_ad_speed );
  SNDLOG"%s\n",s);
  lir_text(2,line,s);
  line++;
  sprintf(s,"No of input bytes = 2 (16 bits)");
  if( (ui.rx_input_mode&DWORD_INPUT) != 0)  
    {
    sprintf(&s[20],"4 (32 bits)");
    }
  SNDLOG"%s\n",s);
  lir_text(2,line,s);
  line++;
  sprintf(s,"Radio interface   =" );
  SNDLOG"%s  ",s);
  lir_text(2,line,s);
  tmptxt="One Rx channel";
  if( (ui.rx_input_mode&TWO_CHANNELS) != 0)tmptxt="Two Rx channels";
  SNDLOG"  %s",tmptxt);
  lir_text(22,line,tmptxt); 
  tmptxt="/ Normal Audio";
  if( (ui.rx_input_mode&IQ_DATA) != 0)tmptxt="/ I/Q ";
  SNDLOG"  %s\n",tmptxt);
  lir_text(38,line,tmptxt);
  line++;
  sprintf(s,
       "(ui.rx_input_mode=%i  ui.rx_rf_channels=%i  ui.rx_ad_channels=%i)", 
                     ui.rx_input_mode, ui.rx_rf_channels, ui.rx_ad_channels );
  SNDLOG"%s  ",s);
  lir_text(2,line,s);
  line++;
  }
line+=2;
// **************************************************************
// Show the current RX output settings on the screen.
// **************************************************************
lir_text(0,line,"Linrad RX output to:");
settextcolor(10);
sprintf(s,"UNKNOWN (no=%d)",ui.rx_dadev_no);
tmptxt=s;
i=0;
k=1;
if(ui.rx_dadev_no == -1)
  {
  tmptxt="NOT YET SELECTED";
  k=0;
  }
else
  {
#ifdef ALSA_PRESENT
  if(ui.use_alsa)
    {
    err=alsa_get_dev_id (ui.rx_dadev_no);
    if(err < 0)
      {
      sprintf(s,"invalid card/device  ( alsa-device-seq-nmbr = %i )",
                                                            ui.rx_dadev_no );
      tmptxt=s;
      k=0;
      }
    else
      {
      sprintf(s,"Soundcard =%s",alsa_dev_soundcard_name);
      lir_text(22,line,s);
      settextcolor(7);
      i++;
      SNDLOG"Initial RXin setting: %s\n",s);
      sprintf(s,"Dev name = %s",alsa_dev_name);
      lir_text(2,line+i,s);
      i++;
      SNDLOG"%s\n",s);
      sprintf(s,"Pcm name = %s",alsa_dev_plughw_pcm_name);
      lir_text(2,line+i,s);
      i++;
      SNDLOG"%s\n",s);
      sprintf(s,"Seq nmbr = %i",alsa_dev_seq_nmbr);
      lir_text(2,line+i,s);
      SNDLOG"%s\n",s);
      }
    }
  else
#endif
    {
    make_devname(ui.rx_dadev_no, dev_name);
    sprintf(s,"Device name =%s",dev_name);
    tmptxt=s;
    }
  }        
if(i == 0)
  {
  lir_text(22,line,tmptxt);
  settextcolor(7);
  SNDLOG"Initial RXin setting: %s\n",tmptxt);
  }
line+=i+1;
if(k)
  {
  sprintf(s,"min da sample rate = %i",ui.rx_min_da_speed);
  SNDLOG"%s\n",s);
  lir_text(2,line,s);
  line++;
  sprintf(s,"max da sample rate = %i ", ui.rx_max_da_speed);
  SNDLOG"%s\n",s);
  lir_text(2,line,s);
  line++;
  sprintf(s,"min da bytes       = %i",ui.rx_min_da_bytes);
  SNDLOG"%s\n",s);
  lir_text(2,line,s);
  line++;
  sprintf(s,"max da bytes       = %i ", ui.rx_max_da_bytes);
  SNDLOG"%s\n",s);
  lir_text(2,line,s);
  line++;
  sprintf(s,"min da channels    = %i",ui.rx_min_da_channels);
  SNDLOG"%s\n",s);
  lir_text(2,line,s);
  line++;
  sprintf(s,"max da channels    = %i", ui.rx_max_da_channels);
  SNDLOG"%s\n",s);
  lir_text(2,line,s);
  line++;
  sprintf(s,"max dma rate       = %i", ui.max_dma_rate);
  SNDLOG"%s\n\n",s);
  lir_text(2,line,s);
  line++;
  }
line+=2;
sndlog_sync();
display_sdi();
// **************************************************************
// Ask whether the user wants to change something and act accordingly.
// **************************************************************
lir_text(0,line,"A = Change the input settings.");
line++;
lir_text(0,line,"B = Change the output settings.");
line++;
lir_text(0,line,"X = To main menu.");
setup_menu:;
refline=0;
toupper_await_keyboard();
if(kill_all_flag)goto set_rx_io_errexit;
line=0;
switch(lir_inkey)
  {
  case 'A':
  clear_screen();
  ui.rx_addev_no=-1;
  if((ui.network_flag&NET_RX_INPUT) != 0)break;
  timf1_char=malloc(0x8000);
  if(timf1_char == NULL)
    {
    lirerr(1231231);
    return;
    }
  init_sdr14();
  free(timf1_char);
  if(kill_all_flag)return;
  if(ui.rx_addev_no == -1)
    {
// look for other hardwares
    }  
  if(ui.rx_addev_no != -1)break;
  sprintf(s,"SELECT SOUNDCARD/DEVICE FOR RX INPUT FROM LIST:");
  SNDLOG"%s\n",s);
  lir_text(0,line,s);
  line++;
#ifdef ALSA_PRESENT
  if(ui.use_alsa)
    {
    ui.rx_addev_no=select_alsadev(SND_PCM_STREAM_CAPTURE, &line);
    if(kill_all_flag)goto set_rx_io_errexit;
    sprintf(s,"Enter sample-rate >");
    lir_text(0,line,s);
enter_sample_rate:;
    SNDLOG"%s\n",s);
    i=lir_get_integer(21, line, 6, 5000,192000);
    if(kill_all_flag)goto set_rx_io_errexit;
    sprintf(s,"User selected  %i",i);
    SNDLOG"%s\n",s);  
    if( (i <= alsa_dev_max_rate) & (i >= alsa_dev_min_rate) )
      {
      ui.rx_ad_speed=i; 
      }
    else
      {
      settextcolor(12);
      sprintf(s,"invalid sample-rate, try again" );
      lir_text(29,line,s);
      SNDLOG"%s\n",s);
      settextcolor(7);
      goto enter_sample_rate;
      }
    lir_text(29,line,blanks);
    line++;
    SNDLOG"\n\n");
//get number of bytes and check
    if(alsa_dev_max_bytes > 2)
      {
      sprintf(s,"Enter number of input bytes (2 or 4) >");
select_number_input_bytes:;
      SNDLOG"%s\n",s);
      lir_text(0,line,s);
      i=lir_get_integer(39, line, 1,2,4);
      if(kill_all_flag)goto set_rx_io_errexit;
      sprintf(s,"User selected  %i",i);
      SNDLOG"%s\n",s); 
      lir_text(44,line,blanks);
      switch (i)
        {
        case 2:
        ui.rx_input_mode=0;
        break;

        case 4:
        ui.rx_input_mode=DWORD_INPUT;
        break;

        default:
        settextcolor(12);
        sprintf(s,"invalid selection, try again");
        SNDLOG"%s\n",s);
        lir_text(44,line,s);
        settextcolor(7);
        goto select_number_input_bytes;
        }
      }
    else
      {
      ui.rx_input_mode=0;
      SNDLOG"\n\n");
      }
    dev_min_channels=1;      
    dev_max_channels=alsa_dev_max_channels;      
    if(dev_max_channels > 4)dev_max_channels=4;
//  With native alsa, always set ui.rx_admode to 0 
    ui.rx_admode=0;
    line+=2;
    }
  else
#endif
    {
    select_ossdev(&line, RXAD);
    max=dev_max_rd_speed[ui.rx_addev_no&255];
    min=dev_min_rd_speed[ui.rx_addev_no&255];
    if(max == ABOVE_MAX_SPEED || max < 1000)
      {
      max=ABOVE_MAX_SPEED;
      speed_warning=1;
      }
    sprintf(s,"Select sampling speed(%d to %d)", min,max);
    lir_text(0,line,s);
    if(speed_warning != 0)
      {
      settextcolor(12);
      lir_text(0,line+2,
       "WARNING!! device does not respond correctly to speed setting command");
      lir_text(0,line+3,
           "Incorrect (unsupported) values may cause system to hang");
      settextcolor(7);
      }
    column=0;
    while(s[column] != 0)column++;
    ui.rx_ad_speed=lir_get_integer(column+2, line, 8, min,max);
    lir_refresh_screen();
    if(kill_all_flag)goto set_rx_io_errexit;
    rdwr_max_speed=ui.rx_ad_speed;
    SNDLOG"\n%d Hz selected by user",ui.rx_ad_speed);
    line++;
    clear_lines(line,line+3);
    if(ioctl(rx_audio_in, SNDCTL_DSP_SPEED, &ui.rx_ad_speed)==-1)
      {
      lirerr(1041);
      goto set_rx_io_errexit;
      }
    if(ui.rx_ad_speed > (int)(rdwr_max_speed*1.05) ||
       ui.rx_ad_speed < (int)(rdwr_max_speed*0.95) )
      {
// The user selected something that was unacceptable to the device driver.
// Rather than just accepting whatever value goto set_rx_io_errexited,
// could be -22 (an unspecified error code) we step the desired speed
// upwards, looking for a valid speed.
// In case we fail, an attempt is made downwards as well.
// ****************************************************************
// This code is written by Ulf, SM0LCB
      SNDLOG"\nReqested speed didn't work");
// This code will search up/down from requested speed
// until the driver will respond with valid code
// Search use delta stepping of 3% from current speed
// This to speed up the search for a valid speed
// First search up in fq
      SNDLOG"\nSearch up : ");
      for( trySpeed=rdwr_max_speed, deltaSpeed=(float)(rdwr_max_speed)*0.03;
                     trySpeed<192000;
                     trySpeed+=deltaSpeed )
        {
        ui.rx_ad_speed=trySpeed;
        if(ioctl(rx_audio_in, SNDCTL_DSP_SPEED, &ui.rx_ad_speed)==-1)
          {
          lirerr(1024);
          goto set_rx_io_errexit;
          }
// test if valid speed
        if(ui.rx_ad_speed < (int)(trySpeed*1.05) &&
           ui.rx_ad_speed > (int)(trySpeed*0.95) )
          {
          SNDLOG"try speed %d Hz ", trySpeed);
          goto spok;
          }
        deltaSpeed=(float)(trySpeed)*0.03;
        }
      SNDLOG"no speed found");
// Then search down in fq
      SNDLOG"\nSearch down : ");
      for( trySpeed=rdwr_max_speed, deltaSpeed=(float)(rdwr_max_speed)*0.03;
                     trySpeed>5000;
                     trySpeed-=deltaSpeed )
        {
        ui.rx_ad_speed=trySpeed;
        if(ioctl(rx_audio_in, SNDCTL_DSP_SPEED, &ui.rx_ad_speed)==-1)
          {
          lirerr(1135);
          goto set_rx_io_errexit;
          }
// test if valid speed
        if(ui.rx_ad_speed < (int)(trySpeed*1.05) &&
           ui.rx_ad_speed > (int)(trySpeed*0.95) )
          {
          SNDLOG"try speed %d Hz ", trySpeed);
          goto spok;
          }
        deltaSpeed=(float)(trySpeed)*0.03;
        }
      SNDLOG"no speed found");
// If we didn't find any vaid speed
      SNDLOG"\nNo valid speed found !");
// Exit program
      lirerr(1134);
      goto set_rx_io_errexit;
      }
spok:;
    close(rx_audio_in);
    rx_audio_in=-1;
    SNDLOG"\n%d Hz reported by device",ui.rx_ad_speed);
    sndlog_sync();
    sprintf(s,"Input speed: %.3fkHz",0.001*(float)(ui.rx_ad_speed));
    lir_text(0,line,s);
    line++;
    if(dev_rd_bytes[ui.rx_addev_no&254] == 6)
      {
      line++;
      lir_text(0,line,"Select A/D number of bytes (2 or 4):");
      line++;
selfmt:;
      await_processed_keyboard();
      if(kill_all_flag) goto set_rx_io_errexit;
      if(lir_inkey == '4')
        {
        ui.rx_input_mode=DWORD_INPUT;
        }
      else
        {
        if(lir_inkey != '2')goto selfmt;
        ui.rx_input_mode=0;
        }
      }
    else
      {
      ui.rx_input_mode=0;
      if(dev_rd_bytes[ui.rx_addev_no&4] == 4)ui.rx_input_mode=DWORD_INPUT;
      }  
    dev_max_channels=0;
    k=4;
    while(dev_max_channels == 0 && k> 0)
      {  
      dev_max_channels|=dev_rd_channels[ui.rx_addev_no]&k;
      k/=2;
      }
    dev_min_channels=0;
    k=1;
    while(dev_min_channels == 0 && k< 8)
      {  
      dev_min_channels|=dev_rd_channels[ui.rx_addev_no]&k;
      k*=2;
      }
    }
  if(ui.rx_addev_no < 256)
    {        
    j=dev_max_channels;
    k=dev_min_channels;
    line++;
    lir_text(0,line,"Select radio interface:");
    line++;
    if(j == 2)j=3;
    for(i=k; i<=j; i++)
      {
      sprintf(s,"%d: %s",i,audiomode_text[i-1]);
      lir_text(5,line+i,s);
      }
chsel:;
    await_processed_keyboard();
    if(kill_all_flag) goto set_rx_io_errexit;
    if(lir_inkey-'0'<k || lir_inkey-'0'>j)goto chsel;
    switch (lir_inkey)
      {
      case '1':
      ui.rx_rf_channels=1;
      ui.rx_ad_channels=1;
      break;

      case '2':
      ui.rx_input_mode|=IQ_DATA;
      ui.rx_rf_channels=1;
      ui.rx_ad_channels=2;
      break;

      case '3':
      ui.rx_input_mode|=TWO_CHANNELS;
      ui.rx_rf_channels=2;
      ui.rx_ad_channels=2;
      break;

      case '4':
      ui.rx_input_mode|=TWO_CHANNELS+IQ_DATA;
      ui.rx_rf_channels=2;
      ui.rx_ad_channels=4;
      break;

      default:
      goto chsel;
      }
    SNDLOG"\nui.rx_input_mode=%d  ui.rx_rf_channels=%d  ui.rx_ad_channels =%d",
                       ui.rx_input_mode, ui.rx_rf_channels, ui.rx_ad_channels);
    if(ui.rx_addev_no == ui.rx_dadev_no)ui.rx_dadev_no=-1;
    if(ui.rx_addev_no == ui.tx_dadev_no)ui.tx_dadev_no=-1;
    if(ui.rx_addev_no == ui.tx_addev_no)ui.tx_addev_no=-1;
    }
  break;
  
  case 'B':
  clear_screen();
  if(ui.rx_addev_no == -1)
    {
    lir_text(0,line,"Set rx input device first.");
    line+=2;
    lir_text(10,line,press_any_key);
    await_keyboard();
    break;
    }
  ui.rx_dadev_no=-1;
  if(ui.rx_addev_no >=0 && ui.rx_addev_no < SPECIFIC_DEVICE_CODES)
    {
    if(start_iotest(&line,RXAD) == FALSE)goto set_rxout_exit;
    }
  sprintf(s,"SELECT SOUNDCARD/DEVICE FOR RX OUTPUT FROM LIST:");
  SNDLOG"%s\n",s);
  lir_text(0,line,s);
  line++;
#ifdef ALSA_PRESENT
  if(ui.use_alsa)
    {
    ui.rx_dadev_no=select_alsadev(SND_PCM_STREAM_PLAYBACK, &line);
    if(kill_all_flag)goto set_rxout_exit;
    ui.rx_min_da_speed=alsa_dev_min_rate;
    ui.rx_max_da_speed=alsa_dev_max_rate;
    if(alsa_dev_min_bytes >=  2 )
      {
      ui.rx_min_da_bytes=2;
      }
    else 
      {
      ui.rx_min_da_bytes=1;
      }
    if(alsa_dev_max_bytes <  2 )
      {
      ui.rx_max_da_bytes=1;
      }
    else
      {
      ui.rx_max_da_bytes=2;
      }
    if(alsa_dev_min_channels >=  2 )ui.rx_min_da_channels=2;
       else ui.rx_min_da_channels=1;
    if(alsa_dev_max_channels <  2 )ui.rx_max_da_channels=1;
       else ui.rx_max_da_channels=2;
    }
  else
#endif
    {
    refline=line;
    select_ossdev(&line, RXDA);
    if(ui.rx_dadev_no == (ui.rx_addev_no&255) && ui.rx_admode == O_RDWR)
      {
      if( (ui.rx_ad_channels == 4) || (ui.rx_input_mode&DWORD_INPUT) != 0)
        {
        lir_text(line,0,
        "RDWR for output not possible with 32 bit data or 4 channels.");
        line++;
        ui.rx_dadev_no=-1;
        lir_text(line,10,press_any_key);
        await_keyboard();
        goto set_rxout_exit;
        }        
      ui.rx_min_da_speed=ui.rx_ad_speed;
      ui.rx_max_da_speed=ui.rx_ad_speed;
      ui.rx_min_da_bytes=2;
      ui.rx_max_da_bytes=2;
      }
    else
      {  
      set_rx_dadev_parms(&line);
      }
    }
// Now that we decided what device to use for output, let
// the user decide how to use it. (It may become microphone input).
  if(ui.rx_dadev_no != -1)
    {
    make_devname(ui.rx_dadev_no, dev_name);
    sndlog_sync();
    ui.rx_damode=select_rdwr_mode(&line, ui.rx_dadev_no, O_WRONLY);
    }
set_rxout_exit:;
  if(ui.rx_addev_no >=0 && ui.rx_addev_no < SPECIFIC_DEVICE_CODES)
    {
    stop_iotest(RXAD);
    if(kill_all_flag)goto set_rx_io_errexit;
    if(!ui.use_alsa)
      {
      if(ui.rx_dadev_no == -1)
        {
        if(line > screen_last_line-6)line=screen_last_line-6;
        lir_text(0,line,
           "No device found for 1/2 byte output while input is open.");
        lir_text(0,line+1,
           "You may use Linrad to watch spectra and record raw data.");
        lir_text(0,line+2,
           "Linrad will now look for devices that can be used to");
        lir_text(0,line+3,
           "output sound while processing recorded data.");
        lir_text(0,line+4,press_any_key);
        SNDLOG"\nNo output device found while input is open.");
        await_processed_keyboard();
        if(kill_all_flag) goto set_rx_io_errexit;
        clear_lines(refline,line+4);
        line=refline;
        select_ossdev(&line, RXDA);
        set_rx_dadev_parms(&line);
        if(ui.rx_dadev_no != -1)ui.rx_input_mode|=NO_DUPLEX;
        }
      }
    }
  line++;  
  sprintf(s,"Set max DMA rate in Hz (10-999)");
  SNDLOG"%s\n",s);
  lir_text(2,line, s);
  i=lir_get_integer(34, line, 3, 10,999);
  if(kill_all_flag)goto set_rx_io_errexit;
  sprintf(s,"User selected  %i",i);
  SNDLOG"%s\n",s);
  ui.max_dma_rate=i;
  if(ui.rx_dadev_no == ui.tx_dadev_no)ui.tx_dadev_no=-1;
  if(ui.rx_dadev_no == ui.tx_addev_no)ui.tx_addev_no=-1;
  break;

  case 'X':
  goto set_rx_io_exit;

  default:
  goto setup_menu;
  }
SNDLOG"goto begin_rx_set_io\n\n");
sndlog_sync();
goto begin_rx_set_io;
set_rx_io_exit:;
SNDLOG"\nNormal end");
set_rx_io_errexit:;
SNDLOG"\n");
fclose(sndlog);
write_log=FALSE;  
}

int get_mic_speed(int *line, int min, int max)
{
int i,mm;
char s[80];
line[0]++;
lir_text(2,line[0],
            "Note that sampling speed should be at least 8 times higher");
line[0]++;
lir_text(2,line[0],"than the highest voice frequency.");
line[0]+=2;
sprintf(s,"Select microphone A/D speed %d to %d",min, max);
lir_text(5,line[0],s);
i=0;
while(s[i] != 0)i++;
mm=lir_get_integer(i+7, line[0], 6, min, max);
line[0]++;
return mm;
}

void select_txout_format(int *line)
{
line[0]++;
lir_text(3,line[0],"Select tx format");
line[0]+=2;
lir_text(7,line[0],"1=Single channel (normal audio)");
line[0]++;
lir_text(7,line[0],"2=Two channels (I and Q, frequency shifted)");
line[0]++;
get_txout_dachan:;
await_keyboard();
if(kill_all_flag)return;
if(lir_inkey=='1')
  {
  ui.tx_da_channels=1;
  }
else
  {
  if(lir_inkey=='2')
    {
    ui.tx_da_channels=2;
    }
  else
    {
    goto get_txout_dachan;
    }
  }
}


void set_tx_io(void)
{  
char *tmptxt;
int dev_min_speed[MAX_DEVNAMES];
int i, j, k, err;
int line;
char s[256];
char *tx_logfile_name="soundboard_tx_init.log";
clear_screen();
if((ui.network_flag&NET_RX_INPUT) != 0)
  {
  lir_text(0,0,"Input from the network is currently selected.");
  goto not_now;
  }
if(ui.rx_ad_channels == 0 ||
   ui.rx_addev_no == -1 ||
   ui.rx_max_da_channels == 0 ||
   ui.rx_dadev_no == -1)
  { 
  lir_text(0,0,"The RX input and output must be set up before the TX.");
not_now:;
  lir_text(0,1,"TX is now disabled.");
  ui.tx_addev_no=-1;
  ui.tx_dadev_no=-1;
  lir_text(10,3,press_any_key);
  await_keyboard();
  return;  
  }
memset(blanks, ' ',58);
blanks[59]=0;
if(ui.tx_ad_channels == 0)ui.tx_addev_no=-1;
if(ui.tx_da_channels == 0)ui.tx_dadev_no=-1;
sndlog=fopen(tx_logfile_name, "w");
if(sndlog == NULL)
  {
  lirerr(1016);
  return;
  }
write_log=TRUE;
begin_tx_set_io:;
clear_screen();
if(kill_all_flag)goto set_tx_io_errexit;
#ifdef ALSA_PRESENT
lir_text(0,0,"The current setting is:");
lir_text(0,1,"Use RX setup menu to toggle between native ALSA and alsa-oss.");
i=ui.use_alsa;
settextcolor(10);
if(ui.use_alsa)
  {
  lir_text(24,0,"native ALSA");
  }
else
  {
  lir_text(24,0,"alsa-oss   ");
  }
sprintf(s,"ALSA library version:%s",SND_LIB_VERSION_STR);
settextcolor(7);  
lir_text(37,0,s);
#else
ui.use_alsa=FALSE;
lir_text(0,0,"This system uses OSS, alsa-oss or legacy sound.");
#endif
clear_lines(1,4);
line=2;
// **************************************************************
// Show the current TX output settings on the screen.
// **************************************************************
lir_text(0,line,"Linrad TX output from:");
settextcolor(10);
sprintf(s,"UNKNOWN (no=%d)",ui.tx_dadev_no);
tmptxt=s;
i=0;
j=1;
if(ui.tx_dadev_no == -1)
  {
  tmptxt="NOT YET SELECTED";
  j=0;
  }
else
  {
#ifdef ALSA_PRESENT
  if(ui.use_alsa)
    {
    err=alsa_get_dev_id (ui.tx_dadev_no);
    if(err < 0)
      {
      sprintf(s,"invalid card/device  ( alsa-device-seq-nmbr = %i )",
                                                            ui.tx_dadev_no );
      tmptxt=s;
      j=0;
      }
    else
      {
      sprintf(s,"Soundcard =%s",alsa_dev_soundcard_name);
      lir_text(22,line,s);
      settextcolor(7);
      i++;
      SNDLOG"Initial RXin setting: %s\n",s);
      sprintf(s,"Dev name = %s",alsa_dev_name);
      lir_text(2,line+i,s);
      i++;
      SNDLOG"%s\n",s);
      sprintf(s,"Pcm name = %s",alsa_dev_plughw_pcm_name);
      lir_text(2,line+i,s);
      i++;
      SNDLOG"%s\n",s);
      sprintf(s,"Seq nmbr = %i",alsa_dev_seq_nmbr);
      lir_text(2,line+i,s);
      SNDLOG"%s\n",s);
      }
    }
  else
#endif
    {
    make_devname(ui.tx_dadev_no, dev_name);
    sprintf(s,"Device name =%s",dev_name);
    tmptxt=s;
    }
  }
if(i == 0)
  {
  lir_text(22,line,tmptxt);
  settextcolor(7);
  SNDLOG"Initial RXin setting: %s\n",tmptxt);
  }
line+=i+1;
if(j)
  {
  sprintf(s,"sample rate        = %i ",ui.tx_da_speed );
  SNDLOG"%s\n",s);
  lir_text(2,line,s);
  line++;
  sprintf(s,"No of output bytes = %d (%d bits)", 
                ui.tx_da_bytes, 8*ui.tx_da_bytes);
  SNDLOG"%s\n",s);
  lir_text(2,line,s);
  line++;
  sprintf(s,"No of output chans = %d", ui.tx_da_channels);
  SNDLOG"%s\n",s);
  lir_text(2,line,s);
  line++;
  }
line+=2;
// **************************************************************
// Show the current TX input settings on the screen.
// **************************************************************
lir_text(0,line,"Linrad TX input to:");
settextcolor(10);
sprintf(s,"UNKNOWN (no=%d)",ui.tx_addev_no);
tmptxt=s;
i=0;
k=1;
if(ui.tx_addev_no == -1)
  {
  tmptxt="NOT YET SELECTED";
  k=0;
  }
else
  {
#ifdef ALSA_PRESENT
  if(ui.use_alsa)
    {
    err=alsa_get_dev_id (ui.tx_addev_no);
    if(err < 0)
      {
      sprintf(s,"invalid card/device  ( alsa-device-seq-nmbr = %i )",
                                                            ui.tx_addev_no );
      tmptxt=s;
      k=0;
      }
    else
      {
      sprintf(s,"Soundcard =%s",alsa_dev_soundcard_name);
      lir_text(22,line,s);
      settextcolor(7);
      i++;
      SNDLOG"Initial RXin setting: %s\n",s);
      sprintf(s,"Dev name = %s",alsa_dev_name);
      lir_text(2,line+i,s);
      i++;
      SNDLOG"%s\n",s);
      sprintf(s,"Pcm name = %s",alsa_dev_plughw_pcm_name);
      lir_text(2,line+i,s);
      i++;
      SNDLOG"%s\n",s);
      sprintf(s,"Seq nmbr = %i",alsa_dev_seq_nmbr);
      lir_text(2,line+i,s);
      SNDLOG"%s\n",s);
      }
    }
  else
#endif
    {
    make_devname(ui.tx_addev_no, dev_name);
    sprintf(s,"Device name =%s",dev_name);
    tmptxt=s;
    }
  }        
if(i == 0)
  {
  lir_text(22,line,tmptxt);
  settextcolor(7);
  SNDLOG"Initial RXin setting: %s\n",tmptxt);
  }
line+=i+1;
if(k)
  {
  sprintf(s,"sample rate        = %i ",ui.tx_ad_speed);
  SNDLOG"%s\n",s);
  lir_text(2,line,s);
  line++;
  sprintf(s,"No of input bytes  = %d (%d bits)", 
                ui.tx_ad_bytes, 8*ui.tx_ad_bytes);
  SNDLOG"%s\n",s);
  lir_text(2,line,s);
  line++;
  sprintf(s,"No of input chans  = %d", ui.tx_ad_channels);
  SNDLOG"%s\n",s);
  lir_text(2,line,s);
  }
line+=2;
sndlog_sync();
// **************************************************************
// Ask whether the user wants to change something and act accordingly.
// **************************************************************
lir_text(0,line,"A = Change the output settings.");
line++;
lir_text(0,line,"B = Change the input settings.");
line++;
lir_text(0,line,"X = To main menu.");
setup_menu:;
toupper_await_keyboard();
if(kill_all_flag)goto set_tx_io_errexit;
line=0;
switch(lir_inkey)
  {
  case 'A':
  clear_screen();
  if(ui.rx_addev_no >=0 && ui.rx_addev_no < SPECIFIC_DEVICE_CODES)
    {
    if(start_iotest(&line,RXAD) == FALSE)goto set_txout_exit;
    }
  if(start_iotest(&line,RXDA) == FALSE)goto set_txout_exit;
  sprintf(s,"SELECT SOUNDCARD/DEVICE FOR TX OUTPUT FROM LIST:");
  SNDLOG"%s\n",s);
  lir_text(0,line,s);
  line++;
#ifdef ALSA_PRESENT
  if(ui.use_alsa)
    {
    ui.tx_dadev_no=select_alsadev(SND_PCM_STREAM_PLAYBACK, &line);
    if(kill_all_flag)goto set_txout_exit;
    if(alsa_dev_min_rate != alsa_dev_max_rate)
      {
      sprintf(s,"Select Tx D/A speed %d to %d",
                                 alsa_dev_min_rate, alsa_dev_max_rate);
      lir_text(5,line,s);
      i=0;
      while(s[i] != 0)i++;
      ui.tx_da_speed=lir_get_integer(i+7, line, 6, 
                                 alsa_dev_min_rate, alsa_dev_max_rate);
      if(kill_all_flag)goto set_txout_exit;      
      line++;
      }
    else
      {  
      ui.tx_da_speed=alsa_dev_min_rate;
      }
    select_txout_format(&line);
    if(kill_all_flag)goto set_txout_exit;
    if(alsa_dev_max_bytes > 2)
      {
      line++;
      lir_text(5,line,"Select tx no of bytes (2 or 4)");
alsa_get_tx_dabytes:;
      ui.tx_da_bytes=lir_get_integer(36,line,2,2,4);
      if(ui.tx_da_bytes==3)goto alsa_get_tx_dabytes;
      }
    else
      {
      ui.tx_da_bytes=2;
      }
    }
  else
#endif
    {
    select_ossdev(&line, TXDA);
    if(ui.tx_dadev_no == (ui.rx_addev_no&255) && ui.rx_admode == O_RDWR)
      {
      if(ui.tx_dadev_no == ui.rx_dadev_no)
        {
        lir_text(line,0,"Device used for RX input and output.");
        goto nodup;
        }
      if(ui.rx_ad_channels == 4)
        {
        lir_text(line,0,"RDWR for TX output not possible with 4 channels.");
nodup:;
        line++;
        ui.tx_dadev_no=-1;
        lir_text(line,10,press_any_key);
        await_keyboard();
        goto set_txout_exit;
        }        
//set speed
      ui.tx_da_speed=ui.rx_ad_speed;
      ui.tx_da_bytes=2;
      if( (ui.rx_input_mode&DWORD_INPUT) != 0)ui.tx_da_bytes=4;
      ui.tx_da_channels=2;
      }
    else
      {
      if(dev_min_wr_speed[ui.tx_dadev_no] != dev_max_wr_speed[ui.tx_dadev_no])
        {
        sprintf(s,"Select Tx D/A speed %d to %d",
          dev_min_wr_speed[ui.tx_dadev_no], dev_max_wr_speed[ui.tx_dadev_no]);
        lir_text(5,line,s);
        i=0;
        while(s[i] != 0)i++;
        ui.tx_da_speed=lir_get_integer(i+7, line, 6, 
          dev_min_wr_speed[ui.tx_dadev_no], dev_max_wr_speed[ui.tx_dadev_no]);
        line++;
        }
      else
        {  
        ui.tx_da_speed=dev_min_speed[ui.tx_dadev_no];
        }
// Select no of channels if the device can do both one and two channels.
      if( (dev_wr_channels[ui.tx_dadev_no]&3) == 3)
        {
        select_txout_format(&line);
        if(kill_all_flag)goto set_txout_exit;
        }
      else
        {
        ui.tx_da_channels=3&dev_wr_channels[ui.tx_dadev_no];
        }
      if( (dev_wr_bytes[ui.tx_dadev_no]&6) == 6)
        {
        line++;
        lir_text(5,line,"Select tx no of bytes (2 or 4)");
get_tx_dabytes:;
        ui.tx_da_bytes=lir_get_integer(36,line,2,2,4);
        if(ui.tx_da_bytes==3)goto get_tx_dabytes;
        }
      else
        {
        ui.tx_da_bytes=dev_wr_bytes[ui.tx_dadev_no]&6;
        }
      }
    }  
set_txout_exit:;
  stop_iotest(RXDA);
  if(ui.rx_addev_no >=0 && ui.rx_addev_no < SPECIFIC_DEVICE_CODES)
    {
    stop_iotest(RXAD);
    }
  if(ui.tx_addev_no == ui.rx_dadev_no)ui.rx_dadev_no=-1;
  break;
  
  case 'B':
  clear_screen();
  if(ui.tx_dadev_no == -1)
    {
    lir_text(0,line,"Set tx output device first.");
    line+=2;
    lir_text(10,line,press_any_key);
    await_keyboard();
    break;
    }
  ui.tx_addev_no=-1;
  if(ui.rx_addev_no >=0 && ui.rx_addev_no < SPECIFIC_DEVICE_CODES)
    {
    if(start_iotest(&line,RXAD) == FALSE)goto set_txin_exit;
    }
  if(start_iotest(&line,RXDA) == FALSE)goto set_txin_exit;
  if(start_iotest(&line,TXDA) == FALSE)goto set_txin_exit;
  sprintf(s,"SELECT SOUNDCARD/DEVICE FOR TX INPUT FROM LIST:");
  SNDLOG"%s\n",s);
  lir_text(0,line,s);
  line++;
#ifdef ALSA_PRESENT
  if(ui.use_alsa)
    {
    ui.tx_addev_no=select_alsadev(SND_PCM_STREAM_CAPTURE, &line);
    if(kill_all_flag)goto set_tx_io_errexit;
    sprintf(s,"Enter sample-rate >");
    lir_text(0,line,s);
enter_sample_rate:;
    SNDLOG"%s\n",s);
    i=lir_get_integer(21, line, 6, 5000,192000);
    if(kill_all_flag)goto set_tx_io_errexit;
    sprintf(s,"User selected  %i",i);
    SNDLOG"%s\n",s);  
    if( (i <= alsa_dev_max_rate) & (i >= alsa_dev_min_rate) )
      {
      ui.tx_ad_speed=i; 
      }
    else
      {
      settextcolor(12);
      sprintf(s,"invalid sample-rate, try again" );
      lir_text(29,line,s);
      SNDLOG"%s\n",s);
      settextcolor(7);
      goto enter_sample_rate;
      }
    lir_text(29,line,blanks);
    line++;
    SNDLOG"\n\n");    
//get number of bytes and check
    if(alsa_dev_max_bytes > 2)
      {
      sprintf(s,"Enter number of input bytes (2 or 4) >");
select_number_input_bytes:;
      SNDLOG"%s\n",s);
      lir_text(0,line,s);
      ui.tx_ad_bytes=lir_get_integer(39, line, 1,2,4);
      if(kill_all_flag)goto set_tx_io_errexit;
      if(ui.tx_ad_bytes==3)
        {
        settextcolor(12);
        sprintf(s,"invalid selection, try again");
        SNDLOG"%s\n",s);
        lir_text(44,line,s);
        settextcolor(7);
        goto select_number_input_bytes;
        }
      line++;
      }
    else
      {
      ui.tx_ad_bytes=2;
      SNDLOG"\n\n");
      }
    ui.tx_ad_channels=1;  
    line+=2;
    }
  else
#endif
    {
    select_ossdev(&line, TXAD);
    if(ui.tx_addev_no == ui.rx_dadev_no && ui.rx_damode == O_RDWR)
      {
      lir_text(0,line,"Rx output parameters modified for RDWR usage");
      line++;
      if(ui.rx_max_da_bytes != 2)
        {
        ui.tx_addev_no=-1;
        goto set_txin_exit;
        }
      ui.rx_min_da_bytes=2;
      ui.tx_ad_bytes=2;  
      if(ui.rx_max_da_channels == 2)
        {
        ui.rx_min_da_channels=2;
        ui.tx_ad_channels=2;
        }
      else
        {
        ui.rx_max_da_channels=1;
        ui.tx_ad_channels=1;
        }
      ui.tx_ad_speed=get_mic_speed(&line, ui.rx_min_da_speed,
                                                        ui.rx_max_da_speed);
      ui.rx_min_da_speed=ui.tx_ad_speed;
      ui.rx_max_da_speed=ui.tx_ad_speed;
      goto set_txin_exit;
      }
// Use one channel if the device can do both one and two channels.
    if( (3 & dev_rd_channels[ui.tx_addev_no]) == 3)
      {
      ui.tx_ad_channels=1;
      }
    else
      {
      ui.tx_ad_channels=3&dev_rd_channels[ui.tx_addev_no];
      }
    if( (dev_rd_channels[ui.tx_addev_no]&3) == 3)
      {
      line++;
      lir_text(0,line,"Select microphone no of bytes (2 or 4)");
get_micbytes:;
      ui.tx_ad_bytes=lir_get_integer(39,line,2,2,4);
      if(ui.tx_ad_bytes==3)goto get_micbytes;
      line++;
      }
    else
      {
      ui.tx_ad_bytes=dev_rd_channels[ui.tx_addev_no]&3;
      }
    if(dev_min_rd_speed[ui.tx_addev_no] != dev_max_rd_speed[ui.tx_addev_no])
      {
      ui.tx_ad_speed=get_mic_speed(&line, dev_min_rd_speed[ui.tx_addev_no], 
                                         dev_max_rd_speed[ui.tx_addev_no]);
      }
    else
      {  
      ui.tx_ad_speed=dev_min_rd_speed[ui.tx_addev_no];
      }
    }  
set_txin_exit:;
  stop_iotest(TXDA);
  stop_iotest(RXDA);
  if(ui.rx_addev_no >=0 && ui.rx_addev_no < SPECIFIC_DEVICE_CODES)
    {
    stop_iotest(RXAD);
    }
  if(ui.tx_addev_no != -1)
    {
    line=0;
    clear_screen();  
get_enable:;
    if(kill_all_flag)break;
    lir_text(4,line,"Enable Tx while running receive modes ? (Y/N)");
    toupper_await_keyboard();
    if(lir_inkey == 'Y')
      {
      ui.tx_enable=1;
      line+=2;
      lir_text(0,line,"Parameters for speech processor, keying, etc. have to");
      line++;
      lir_text(0,line,"be set up and saved before Tx becomes enabled in Rx modes.");
      line++;
      }
    else
      {
      if(lir_inkey != 'N')goto get_enable;  
      ui.tx_enable=0;
      }  
    line+=2;
    lir_text(0,line,"The tx pilot tone may control fast Rx/Tx switching.");
    line++;
    lir_text(0,line,"Set pilot tone level in dB below signal (50-130)"); 
    ui.tx_pilot_tone_db=lir_get_integer(49, line, 3, 50, 130);
    line+=2;
    lir_text(0,line,"Set prestart in microseconds (0-100000)");
    ui.tx_pilot_tone_prestart=lir_get_integer(39, line, 6, 0, 100000);
    line+=2;
    settextcolor(15);
    lir_text(5,line,remind_parsave);
    settextcolor(7);
    line+=2;
    lir_text(5,line,press_any_key);
    await_keyboard();
    }
  break;

  case 'X':
  goto set_tx_io_exit;

  default:
  goto setup_menu;
  }  
SNDLOG"goto begin_tx_set_io\n\n");
sndlog_sync();
goto begin_tx_set_io;
set_tx_io_exit:;
SNDLOG"\nNormal end");
set_tx_io_errexit:;
SNDLOG"\n");
fclose(sndlog);
write_log=FALSE;
}

